@@ -337,7 +337,10 @@ store_fixed_bit_field (op0, offset, bitsize, bitpos, value)
     subtarget = expand_bit_and (mode, op0,
 				gen_rtx (CONST_INT, VOIDmode, 
 					 (~ (((1 << bitsize) - 1) << bitpos))
-					  & ((1 << GET_MODE_BITSIZE (mode)) - 1)),
+					 & ((GET_MODE_BITSIZE (mode)
+					     == HOST_BITS_PER_INT)
+					    ? -1
+					    : ((1 << GET_MODE_BITSIZE (mode)) - 1))),
 				subtarget);
 
   /* Now logical-or VALUE into OP0, unless it is zero.  */
@@ -537,8 +540,27 @@ extract_bit_field (str_rtx, bitsize, bitnum, unsignedp, target, mode, tmode)
 	{
 	  /* Get ref to first byte containing part of the field.  */
 	  if (GET_CODE (op0) == MEM)
-	    op0 = change_address (op0, QImode,
-				  plus_constant (XEXP (op0, 0), offset));
+	    {
+	      if (! ((*insn_operand_predicate[(int) CODE_FOR_extzv][1])
+		     (op0, GET_MODE (op0))))
+		{
+		  /* If memory isn't acceptable for this operand,
+		     copy it to a register.  */
+		  unit = BITS_PER_WORD;
+		  offset = bitnum / unit;
+		  bitpos = bitnum % unit;
+		  op0 = change_address (op0, SImode,
+					plus_constant (XEXP (op0, 0), offset));
+		  op0 = force_reg (GET_MODE (op0), op0);
+#ifdef BITS_BIG_ENDIAN
+		  if (unit > GET_MODE_BITSIZE (GET_MODE (op0)))
+		    bitpos += unit - GET_MODE_BITSIZE (GET_MODE (op0));
+#endif
+		}
+	      else
+		op0 = change_address (op0, QImode,
+				      plus_constant (XEXP (op0, 0), offset));
+	    }
 
 	  /* If op0 is a register, we need it in SImode
 	     to make it acceptable to the format of extv.  */
@@ -588,8 +610,28 @@ extract_bit_field (str_rtx, bitsize, bitnum, unsignedp, target, mode, tmode)
 	{
 	  /* Get ref to first byte containing part of the field.  */
 	  if (GET_CODE (op0) == MEM)
-	    op0 = change_address (op0, QImode,
-				  plus_constant (XEXP (op0, 0), offset));
+	    {
+	      if (! ((*insn_operand_predicate[(int) CODE_FOR_extzv][1])
+		     (op0, GET_MODE (op0))))
+		{
+		  /* If memory isn't acceptable for this operand,
+		     copy it to a register.  */
+		  unit = BITS_PER_WORD;
+		  offset = bitnum / unit;
+		  bitpos = bitnum % unit;
+		  op0 = change_address (op0, SImode,
+					plus_constant (XEXP (op0, 0), offset));
+		  op0 = force_reg (GET_MODE (op0), op0);
+#ifdef BITS_BIG_ENDIAN
+		  if (unit > GET_MODE_BITSIZE (GET_MODE (op0)))
+		    bitpos += unit - GET_MODE_BITSIZE (GET_MODE (op0));
+#endif
+		}
+	      else
+		op0 = change_address (op0, QImode,
+				      plus_constant (XEXP (op0, 0), offset));
+	    }
+
 
 	  /* If op0 is a register, we need it in QImode
 	     to make it acceptable to the format of extv.  */
@@ -728,9 +770,9 @@ extract_fixed_bit_field (tmode, op0, offset, bitsize, bitpos, target, unsignedp)
 	     shift it so it does.  */
 	  tree amount = build_int_2 (bitpos, 0);
 	  /* Maybe propagate the target for the shift.  */
-	  /* Certainly do so if we will return the value of the shift.  */
+	  /* But not if we will return it--could confuse integrate.c.  */
 	  rtx subtarget = (target != 0 && GET_CODE (target) == REG
-			   && FUNCTION_VALUE_REGNO_P (REGNO (target))
+			   && !REG_FUNCTION_VALUE_P (target)
 			   ? target : 0);
 	  if (tmode != mode) subtarget = 0;
 	  op0 = expand_shift (RSHIFT_EXPR, mode, op0, amount, subtarget, 1);
@@ -774,9 +816,9 @@ extract_fixed_bit_field (tmode, op0, offset, bitsize, bitpos, target, unsignedp)
     {
       tree amount = build_int_2 (GET_MODE_BITSIZE (mode) - (bitsize + bitpos), 0);
       /* Maybe propagate the target for the shift.  */
-      /* Certainly do so if we will return the value of the shift.  */
+      /* But not if we will return the result--could confuse integrate.c.  */
       rtx subtarget = (target != 0 && GET_CODE (target) == REG
-		       && FUNCTION_VALUE_REGNO_P (REGNO (target))
+		       && ! REG_FUNCTION_VALUE_P (target)
 		       ? target : 0);
       op0 = expand_shift (LSHIFT_EXPR, mode, op0, amount, subtarget, 1);
     }
@@ -1111,7 +1153,7 @@ expand_mult (mode, op0, op1, target, unsignedp)
   /* Don't use the function value register as a target
      since we have to read it as well as write it,
      and function-inlining gets confused by this.  */
-  if (target && REG_P (target) && FUNCTION_VALUE_REGNO_P (REGNO (target)))
+  if (target && REG_P (target) && REG_FUNCTION_VALUE_P (target))
     target = 0;
 
   if (GET_CODE (op1) == CONST_INT)
@@ -1121,6 +1163,9 @@ expand_mult (mode, op0, op1, target, unsignedp)
       int negate = INTVAL (op1) < 0;
       int absval = INTVAL (op1) * (negate ? -1 : 1);
 
+      if (absval == 1)
+	return negate ? negate_rtx (op0) : op0;
+
       /* Is multiplier a power of 2, or minus that?  */
       foo = exact_log2 (absval);
       if (foo >= 0)
@@ -1191,13 +1236,11 @@ expand_divmod (rem_flag, code, mode, op0, op1, target, unsignedp)
   /* Don't use the function value register as a target
      since we have to read it as well as write it,
      and function-inlining gets confused by this.  */
-  if (target && REG_P (target) && FUNCTION_VALUE_REGNO_P (REGNO (target)))
+  if (target && REG_P (target) && REG_FUNCTION_VALUE_P (target))
     target = 0;
 
   if (target == 0)
-    {
-      target = gen_reg_rtx (mode);
-    }
+    target = gen_reg_rtx (mode);
 
   /* Don't clobber an operand while doing a multi-step calculation.  */
   if ((rem_flag && rtx_equal_p (target, op0))
