@@ -99,7 +99,9 @@ force_fit_type (t)
 
   /* First clear all bits that are beyond the type's precision.  */
 
-  if (prec > HOST_BITS_PER_INT)
+  if (prec == 2 * HOST_BITS_PER_INT)
+    ;
+  else if (prec > HOST_BITS_PER_INT)
     {
       TREE_INT_CST_HIGH (t)
 	&= ~((-1) << (prec - HOST_BITS_PER_INT));
@@ -115,6 +117,7 @@ force_fit_type (t)
   /* If it's a signed type and value's sign bit is set, extend the sign.  */
 
   if (! TREE_UNSIGNED (TREE_TYPE (t))
+      && prec != 2 * HOST_BITS_PER_INT
       && (prec > HOST_BITS_PER_INT
 	  ? TREE_INT_CST_HIGH (t) & (1 << (prec - HOST_BITS_PER_INT - 1))
 	  : TREE_INT_CST_LOW (t) & (1 << (prec - 1))))
@@ -840,46 +843,41 @@ combine (code, arg1, arg2)
     {
       register double d1 = TREE_REAL_CST (arg1);
       register double d2 = TREE_REAL_CST (arg2);
-      register tree t;
+      register double value;
 
       switch (code)
 	{
 	case PLUS_EXPR:
-	  t = build_real (d1 + d2);
+	  value = d1 + d2;
 	  break;
 
 	case MINUS_EXPR:
-	  t = build_real (d1 - d2);
+	  value = d1 - d2;
 	  break;
 
 	case MULT_EXPR:
-	  t = build_real (d1 * d2);
+	  value = d1 * d2;
 	  break;
 
 	case RDIV_EXPR:
 	  if (d2 == 0)
 	    return 0;
 
-	  t = build_real (d1 / d2);
+	  value = d1 / d2;
 	  break;
 
 	case MIN_EXPR:
-	  if (d1 < d2)
-	    t = build_real (d1);
-	  t = build_real (d2);
+	  value = d1 < d2 ? d1 : d2;
 	  break;
 
 	case MAX_EXPR:
-	  if (d1 > d2)
-	    t = build_real (d1);
-	  t = build_real (d2);
+	  value = d1 > d2 ? d1 : d2;
 	  break;
 
 	default:
 	  abort ();
 	}
-      TREE_TYPE (t) = TREE_TYPE (arg1);
-      return t;
+      return build_real (TREE_TYPE (arg1), value);
     }
   if (TREE_CODE (arg1) == COMPLEX_CST)
     {
@@ -956,18 +954,6 @@ fold_convert (t)
 	{
 	  /* Given an integer constant, make new constant with new type,
 	     appropriately sign-extended or truncated.  */
-	  register int inprec;
-	  register int outprec;
-
-	  if (TREE_CODE (TREE_TYPE (arg1)) == POINTER_TYPE)
-	    inprec = BITS_PER_WORD;
-	  else
-	    inprec = TYPE_PRECISION (TREE_TYPE (arg1));
-	  if (TREE_CODE (type) == POINTER_TYPE)
-	    outprec = BITS_PER_WORD;
-	  else
-	    outprec = TYPE_PRECISION (type);
-
 	  t = build_int_2 (TREE_INT_CST_LOW (arg1),
 			   TREE_INT_CST_HIGH (arg1));
 	  TREE_TYPE (t) = type;
@@ -976,15 +962,15 @@ fold_convert (t)
       else if (TREE_CODE (arg1) == REAL_CST)
 	t = build_int_2 ((int) TREE_REAL_CST (arg1),
 			 (int) (TREE_REAL_CST (arg1) / 0x10000 / 0x10000));
+      TREE_TYPE (t) = type;
     }
   else if (TREE_CODE (type) == REAL_TYPE)
     {
       if (TREE_CODE (arg1) == INTEGER_CST)
-	t = build_real_from_int_cst (arg1);
+	return build_real_from_int_cst (type, arg1);
       else if (TREE_CODE (arg1) == REAL_CST)
-	t = build_real (TREE_REAL_CST (arg1));
+	return build_real (type, TREE_REAL_CST (arg1));
     }
-  TREE_TYPE (t) = type;
   TREE_LITERAL (t) = 1;
   return t;
 }
@@ -1115,10 +1101,11 @@ fold (expr)
 	      else
 		t = build_int_2 (- TREE_INT_CST_LOW (arg0),
 				 ~ TREE_INT_CST_HIGH (arg0));
+	      TREE_TYPE (t) = TREE_TYPE (expr);
 	      force_fit_type (t);
 	    }
 	  else if (TREE_CODE (arg0) == REAL_CST)
-	    t = build_real (- TREE_REAL_CST (arg0));
+	    t = build_real (TREE_TYPE (expr), - TREE_REAL_CST (arg0));
 	  TREE_TYPE (t) = TREE_TYPE (expr);
 	}
       return t;
@@ -1141,7 +1128,7 @@ fold (expr)
 	  else if (TREE_CODE (arg0) == REAL_CST)
 	    {
 	      if (TREE_REAL_CST (arg0) < 0)
-		t = build_real (- TREE_REAL_CST (arg0));
+		t = build_real (TREE_TYPE (expr), - TREE_REAL_CST (arg0));
 	    }
 	  TREE_TYPE (t) = TREE_TYPE (expr);
 	}
@@ -1185,7 +1172,7 @@ fold (expr)
 		  /* EXPR is (CON-VAR) +- ARG1.  */
 		  /* If it is + and VAR==ARG1, return just CONST.  */
 		  if (code == PLUS_EXPR && operand_equal_p (var, arg1))
-		    return con;
+		    return convert (TREE_TYPE (t), con);
 		    
 		  /* Otherwise return (CON +- ARG1) - VAR.  */
 		  TREE_SET_CODE (t, MINUS_EXPR);
@@ -1198,7 +1185,7 @@ fold (expr)
 		  /* EXPR is (VAR+CON) +- ARG1.  */
 		  /* If it is - and VAR==ARG1, return just CONST.  */
 		  if (code == MINUS_EXPR && operand_equal_p (var, arg1))
-		    return con;
+		    return convert (TREE_TYPE (t), con);
 		    
 		  /* Otherwise return VAR +- (ARG1 +- CON).  */
 		  TREE_OPERAND (t, 1) = tem
@@ -1231,13 +1218,13 @@ fold (expr)
 		TREE_SET_CODE (t,
 			       (code == PLUS_EXPR ? MINUS_EXPR : PLUS_EXPR));
 	      if (TREE_CODE (t) == MINUS_EXPR && operand_equal_p (var, arg0))
-		return con;
+		return convert (TREE_TYPE (t), con);
 	      TREE_OPERAND (t, 0)
 		= fold (build (code, TREE_TYPE (t), arg0, con));
 	      TREE_OPERAND (t, 1) = var;
 	      if (integer_zerop (TREE_OPERAND (t, 0))
 		  && TREE_CODE (t) == PLUS_EXPR)
-		return var;
+		return convert (TREE_TYPE (t), var);
 	      return t;
 	    }
 	}
@@ -1407,6 +1394,32 @@ fold (expr)
     case GT_EXPR:
     case LE_EXPR:
     case GE_EXPR:
+      /* If one arg is a constant integer, put it last.  */
+      if (TREE_CODE (arg0) == INTEGER_CST
+	  && TREE_CODE (arg1) != INTEGER_CST)
+	{
+	  TREE_OPERAND (t, 0) = arg1;
+	  TREE_OPERAND (t, 1) = arg0;
+	  arg0 = TREE_OPERAND (t, 0);
+	  arg1 = TREE_OPERAND (t, 1);
+	  switch (code)
+	    {
+	    case GT_EXPR:
+	      code = LT_EXPR;
+	      break;
+	    case GE_EXPR:
+	      code = LE_EXPR;
+	      break;
+	    case LT_EXPR:
+	      code = GT_EXPR;
+	      break;
+	    case LE_EXPR:
+	      code = GE_EXPR;
+	      break;
+	    }
+	  TREE_CODE (t) = code;
+	}
+
       /* Convert foo++ == CONST into ++foo == CONST + INCR.
 	 First, see if one arg is constant; find the constant arg
 	 and the other one.  */
@@ -1452,6 +1465,48 @@ fold (expr)
 	  }
       }
 
+      /* Change X >= CST to X > (CST - 1) if CST is positive.  */
+      if (TREE_CODE (arg1) == INTEGER_CST
+	  && ! tree_int_cst_lt (arg1, integer_one_node))
+	{
+	  switch (TREE_CODE (t))
+	    {
+	    case GE_EXPR:
+	      TREE_CODE (t) = GT_EXPR;
+	      arg1 = combine (MINUS_EXPR, arg1, integer_one_node);
+	      TREE_OPERAND (t, 1) = arg1;
+	      break;
+
+	    case LT_EXPR:
+	      TREE_CODE (t) = LE_EXPR;
+	      arg1 = combine (MINUS_EXPR, arg1, integer_one_node);
+	      TREE_OPERAND (t, 1) = arg1;
+	    }
+	}
+
+      /* An unsigned comparison against 0 can be simplified.  */
+      if (integer_zerop (arg1)
+	  && (TREE_CODE (TREE_TYPE (arg1)) == INTEGER_TYPE
+	      || TREE_CODE (TREE_TYPE (arg1)) == POINTER_TYPE)
+	  && TREE_UNSIGNED (TREE_TYPE (arg1)))
+	{
+	  switch (TREE_CODE (t))
+	    {
+	    case GT_EXPR:
+	      TREE_CODE (t) = NE_EXPR;
+	      break;
+	    case LE_EXPR:
+	      TREE_CODE (t) = EQ_EXPR;
+	      break;
+	    case GE_EXPR:
+	      return build (COMPOUND_EXPR, integer_type_node,
+			    arg0, integer_one_node);
+	    case LT_EXPR:
+	      return build (COMPOUND_EXPR, integer_type_node,
+			    arg0, integer_zero_node);
+	    }
+	}
+
       /* To compute GT, swap the arguments and do LT.
 	 To compute GE, do LT and invert the result.
 	 To compute LE, swap the arguments, do LT and invert the result.
@@ -1462,7 +1517,9 @@ fold (expr)
 	  arg0 = arg1;
 	  arg1 = temp;
 	}
-      /* Compute a result for LT or EQ if args permit.  */
+
+      /* Compute a result for LT or EQ if args permit;
+	 otherwise return T.  */
       if (TREE_CODE (arg0) == INTEGER_CST
 	  && TREE_CODE (arg1) == INTEGER_CST)
 	{
@@ -1477,6 +1534,13 @@ fold (expr)
 			      : INT_CST_LT (arg0, arg1)),
 			     0);
 	}
+      else if (TREE_CODE (arg1) == INTEGER_CST
+	       && TREE_LITERAL (arg0)
+	       && TREE_CODE (arg0) == ADDR_EXPR
+	       && (code == EQ_EXPR || code == NE_EXPR))
+	{
+	  t = build_int_2 (0, 0);
+	}
       else if (TREE_CODE (arg0) == REAL_CST
 	       && TREE_CODE (arg1) == REAL_CST)
 	{
@@ -1488,7 +1552,7 @@ fold (expr)
       else
 	return t;
 
-      /* If we wanted ...-or-equal, invert the result.  */
+      /* If what we want is other than LT or EQ, invert the result.  */
       if (code == GE_EXPR || code == LE_EXPR || code == NE_EXPR)
 	TREE_INT_CST_LOW (t) ^= 1;
       TREE_TYPE (t) = TREE_TYPE (expr);
