@@ -134,6 +134,10 @@ static HARD_REG_SET *hard_reg_conflicts;
 static HARD_REG_SET *hard_reg_preferences;
 #endif
 
+/* Set of registers that global-alloc isn't supposed to use.  */
+
+static HARD_REG_SET no_global_alloc_regs;
+
 /* Test a bit in TABLE, a vector of HARD_REG_SETs,
    for vector element I, and hard register number J.  */
 
@@ -185,6 +189,16 @@ global_alloc (file)
 
   max_allocno = 0;
 
+  /* A machine may have certain hard registers that
+     are safe to use only within a basic block.  */
+
+  CLEAR_HARD_REG_SET (no_global_alloc_regs);
+#ifdef OVERLAPPING_REGNO_P
+  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
+    if (OVERLAPPING_REGNO_P (i))
+      SET_HARD_REG_BIT (no_global_alloc_regs, i);
+#endif
+
   /* Establish mappings from register number to allocation number
      and vice versa.  In the process, count the allocnos.  */
 
@@ -310,16 +324,19 @@ allocno_compare (v1, v2)
 {
   register int r1 = allocno_reg[*v1];
   register int r2 = allocno_reg[*v2];
-  register double v 
-    = ((double) (floor_log2 (reg_n_refs[r1]) * reg_n_refs[r1])
-       / (reg_live_length[r1] * allocno_size[*v1]))
-      - ((double) (floor_log2 (reg_n_refs[r2]) * reg_n_refs[r2])
-	 / (reg_live_length[r2] * allocno_size[*v2]));
-  if (v < 0)
-    return 1;
-  if (v > 0)
-    return -1;
-  return 0;
+  /* Note that the quotient will never be bigger than
+     the value of floor_log2 times the maximum number of
+     times a register can occur in one insn (surely less than 100).
+     Multiplying this by 10000 can't overflow.  */
+  register int pri1
+    = (((double) (floor_log2 (reg_n_refs[r1]) * reg_n_refs[r1])
+	/ (reg_live_length[r1] * allocno_size[*v1]))
+       * 10000);
+  register int pri2
+    = (((double) (floor_log2 (reg_n_refs[r2]) * reg_n_refs[r2])
+	/ (reg_live_length[r2] * allocno_size[*v2]))
+       * 10000);
+  return pri2 - pri1;
 }
 
 /* Scan the rtl code and record all conflicts in the conflict matrices.  */
@@ -397,7 +414,7 @@ global_conflicts ()
 	{
 	  register RTX_CODE code = GET_CODE (insn);
 	  register rtx link;
-	  rtx regs_set[MAX_SETS_PER_INSN + MAX_CLOBBERS_PER_INSN];
+	  rtx *regs_set = (rtx *) alloca (max_parallel * sizeof (rtx));
 	  int n_regs_set = 0;
 
 	  if (code == INSN || code == CALL_INSN || code == JUMP_INSN)
@@ -509,6 +526,9 @@ find_reg (allocno, losers, all_regs_p, prefreg)
 		     (reg_crosses_call[allocno_reg[allocno]]
 		      ? call_used_reg_set : fixed_reg_set));
 
+  /* Some registers should not be allocated in global-alloc.  */
+  IOR_HARD_REG_SET (used, no_global_alloc_regs);
+
   IOR_COMPL_HARD_REG_SET (used, reg_class_contents[(int) class]);
   IOR_HARD_REG_SET (used, hard_reg_conflicts[allocno]);
   if (frame_pointer_needed)
@@ -537,21 +557,33 @@ find_reg (allocno, losers, all_regs_p, prefreg)
 
   if (i < 0)
     for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
-      if (! TEST_HARD_REG_BIT (used, i)
-	  && (losers == 0 || losers[i] < 0)
-	  && HARD_REGNO_MODE_OK (i, mode))
-	{
-	  register int j;
-	  register int lim = i + HARD_REGNO_NREGS (i, mode);
-	  for (j = i + 1;
-	       (j < lim
-		&& ! TEST_HARD_REG_BIT (used, j)
-		&& (losers == 0 || losers[j] < 0));
-	       j++);
-	  if (j == lim)
-	    break;
-	  i = j;			/* Skip starting points we know will lose */
-	}
+      {
+#ifdef REG_ALLOC_ORDER
+	int regno = reg_alloc_order[i];
+#else
+	int regno = i;
+#endif
+	if (! TEST_HARD_REG_BIT (used, regno)
+	    && (losers == 0 || losers[regno] < 0)
+	    && HARD_REGNO_MODE_OK (regno, mode))
+	  {
+	    register int j;
+	    register int lim = regno + HARD_REGNO_NREGS (regno, mode);
+	    for (j = regno + 1;
+		 (j < lim
+		  && ! TEST_HARD_REG_BIT (used, j)
+		  && (losers == 0 || losers[j] < 0));
+		 j++);
+	    if (j == lim)
+	      {
+		i = regno;
+		break;
+	      }
+#ifndef REG_ALLOC_ORDER
+	    i = j;			/* Skip starting points we know will lose */
+#endif
+	  }
+      }
 
   /* Did we find a register?  */
 
@@ -842,8 +874,7 @@ dump_conflicts (file)
   for (i = 0; i < max_allocno; i++)
     {
       register int j;
-      fprintf (file, ";; %d conflicts:", allocno_reg[i],
-	       reg_renumber[allocno_reg[i]]);
+      fprintf (file, ";; %d conflicts:", allocno_reg[i]);
       for (j = 0; j < max_allocno; j++)
 	if (CONFLICTP (i, j) || CONFLICTP (j, i))
 	  fprintf (file, " %d", allocno_reg[j]);
