@@ -42,6 +42,13 @@ extern struct obstack *rtl_obstack, *saveable_obstack, *current_obstack;
 #define MIN(x,y) ((x < y) ? x : y)
 
 extern tree pushdecl ();
+
+/* Default max number of insns a function can have and still be inline.
+   This is overridden on RISC machines.  */
+#ifndef INTEGRATE_THRESHOLD
+#define INTEGRATE_THRESHOLD(DECL) \
+  (8 * (8 + list_length (DECL_ARGUMENTS (DECL)) + 16*TREE_INLINE (DECL)))
+#endif
 
 /* This is the target of the inline function being expanded,
    or NULL if there is none.  */
@@ -77,6 +84,8 @@ static int fp_delta;
 /* Return a copy of an rtx (as needed), substituting pseudo-register,
    labels, and frame-pointer offsets as necessary.  */
 static rtx copy_rtx_and_substitute ();
+/* Variant, used for memory addresses that are not memory_address_p.  */
+static rtx copy_address ();
 
 static void copy_parm_decls ();
 static void copy_decl_tree ();
@@ -85,7 +94,7 @@ static rtx try_fold_cc0 ();
 
 /* We do some simple constant folding optimization.  This optimization
    really exists primarily to save time inlining a function.  It
-   also help users who ask for inline functions without -O.  */
+   also helps users who ask for inline functions without -O.  */
 static rtx fold_out_const_cc0 ();
 
 /* Zero if the current function (whose FUNCTION_DECL is FNDECL)
@@ -99,8 +108,7 @@ function_cannot_inline_p (fndecl)
 {
   register rtx insn;
   tree last = tree_last (TYPE_ARG_TYPES (TREE_TYPE (fndecl)));
-  int nargs = list_length (DECL_ARGUMENTS (fndecl));
-  int max_insns = 4 * (4 + nargs + 16*TREE_INLINE (fndecl));
+  int max_insns = INTEGRATE_THRESHOLD (fndecl);
   register int ninsns = 0;
   register tree parms;
 
@@ -108,11 +116,11 @@ function_cannot_inline_p (fndecl)
      message about that if `inline' is specified.  This code
      it put in to catch the volunteers.  */
   if (last && TREE_VALUE (last) != void_type_node)
-    return "varargs function `%s' cannot be inline";
+    return "varargs function cannot be inline";
 
   /* If its not even close, don't even look.  */
   if (get_max_uid () > 2 * max_insns)
-    return "function `%s' too large to be inline";
+    return "function too large to be inline";
 
   /* Don't inline functions which have BLKmode arguments.
      Don't inline functions that take the address of
@@ -120,9 +128,9 @@ function_cannot_inline_p (fndecl)
   for (parms = DECL_ARGUMENTS (fndecl); parms; parms = TREE_CHAIN (parms))
     {
       if (TYPE_MODE (TREE_TYPE (parms)) == BLKmode)
-	return "function `%s' with large aggregate parameter cannot be inline";
+	return "function with large aggregate parameter cannot be inline";
       if (last == NULL_TREE && TREE_ADDRESSABLE (parms))
-	return "function `%s' without prototype uses address of parameter;\n cannot be inline";
+	return "no prototype, and parameter address used; cannot be inline";
     }
 
   if (get_max_uid () > max_insns)
@@ -137,7 +145,7 @@ function_cannot_inline_p (fndecl)
 	}
 
       if (ninsns >= max_insns)
-	return "function `%s' too large to be inline";
+	return "function too large to be inline";
     }
 
   return 0;
@@ -147,16 +155,21 @@ function_cannot_inline_p (fndecl)
 
 /* Mapping from old pesudo-register to new pseudo-registers.
    The first element of this map is reg_map[FIRST_PSEUDO_REGISTER].
-   It allocated in `save_current_insns' and `expand_function_inline',
+   It is allocated in `save_for_inline' and `expand_inline_function',
    and deallocated on exit from each of those routines.  */
 static rtx *reg_map;
 
 /* Mapping from old code-labels to new code-labels.
    The first element of this map is label_map[min_labelno].
-   It allocated in `save_current_insns' and `expand_function_inline',
+   It is allocated in `save_for_inline' and `expand_inline_function',
    and deallocated on exit from each of those routines.  */
 static rtx *label_map;
 
+/* Mapping from old insn uid's to copied insns.
+   It is allocated in `save_for_inline' and `expand_inline_function',
+   and deallocated on exit from each of those routines.  */
+static rtx *insn_map;
+
 /* Map pseudo reg number into the PARM_DECL for the parm living in the reg.
    Zero for a reg that isn't a parm's home.
    Only reg numbers less than max_parm_reg are mapped here.  */
@@ -165,6 +178,9 @@ static tree *parmdecl_map;
 /* Keep track of first pseudo-register beyond those that are parms.  */
 static int max_parm_reg;
 
+/* Offset from arg ptr to the first parm of this inline function.  */
+static int first_parm_offset;
+
 /* On machines that perform a function return with a single
    instruction, such as the VAX, these return insns must be
    mapped into branch statements.  */
@@ -189,6 +205,7 @@ save_for_inline (fndecl)
   tree parms;
   int max_labelno, min_labelno, i, len;
   int max_reg;
+  int max_uid;
 
   /* Make and emit a return-label if we have not already done so.  */
 
@@ -211,6 +228,7 @@ save_for_inline (fndecl)
      Later we set TREE_READONLY to 0 if the parm is modified inside the fn.  */
 
   parmdecl_map = (tree *) alloca (max_parm_reg * sizeof (tree));
+  bzero (parmdecl_map, max_parm_reg * sizeof (tree));
 
   for (parms = DECL_ARGUMENTS (fndecl); parms; parms = TREE_CHAIN (parms))
     {
@@ -238,6 +256,7 @@ save_for_inline (fndecl)
   head = gen_inline_header_rtx (NULL, NULL, min_labelno, max_labelno,
 				max_parm_reg, max_reg,
 				current_function_args_size);
+  max_uid = INSN_UID (head);
 
   /* We have now allocated all that needs to be allocated permanently
      on the rtx obstack.  Set our high-water mark, so that we
@@ -277,7 +296,7 @@ save_for_inline (fndecl)
     reg_map[i] = (rtx)obstack_copy (&maybepermanent_obstack, regno_reg_rtx[i], len);
   bcopy (reg_map + FIRST_PSEUDO_REGISTER,
 	 regno_reg_rtx + FIRST_PSEUDO_REGISTER,
-	 (max_reg_num () - FIRST_PSEUDO_REGISTER) * sizeof (rtx));
+	 (max_reg - FIRST_PSEUDO_REGISTER) * sizeof (rtx));
 
   /* Likewise each label rtx must have a unique rtx as its copy.  */
 
@@ -287,6 +306,11 @@ save_for_inline (fndecl)
   for (i = min_labelno; i < max_labelno; i++)
     label_map[i] = gen_label_rtx ();
 
+  /* Record the mapping of old insns to copied insns.  */
+
+  insn_map = (rtx *) alloca (max_uid * sizeof (rtx));
+  bzero (insn_map, max_uid * sizeof (rtx));
+
   /* Now copy the chain of insns.  */
 
   for (insn = NEXT_INSN (insn); insn; insn = NEXT_INSN (insn))
@@ -294,6 +318,9 @@ save_for_inline (fndecl)
       switch (GET_CODE (insn))
 	{
 	case NOTE:
+	  if (NOTE_LINE_NUMBER (insn) == NOTE_INSN_FUNCTION_END)
+	    continue;
+
 	  copy = rtx_alloc (NOTE);
 	  NOTE_SOURCE_FILE (copy) = NOTE_SOURCE_FILE (insn);
 	  NOTE_LINE_NUMBER (copy) = NOTE_LINE_NUMBER (insn);
@@ -321,6 +348,7 @@ save_for_inline (fndecl)
 	  abort ();
 	}
       INSN_UID (copy) = INSN_UID (insn);
+      insn_map[INSN_UID (insn)] = copy;
       NEXT_INSN (last_insn) = copy;
       PREV_INSN (copy) = last_insn;
       last_insn = copy;
@@ -369,7 +397,6 @@ copy_for_inline (orig)
     case CONST_INT:
     case CONST_DOUBLE:
     case SYMBOL_REF:
-    case CODE_LABEL:
     case PC:
     case CC0:
       return x;
@@ -412,7 +439,8 @@ copy_for_inline (orig)
 
 	if (GET_CODE (dest) == REG
 	    && REGNO (dest) < max_parm_reg
-	    && REGNO (dest) >= FIRST_PSEUDO_REGISTER)
+	    && REGNO (dest) >= FIRST_PSEUDO_REGISTER
+	    && parmdecl_map[REGNO (dest)] != 0)
 	  TREE_READONLY (parmdecl_map[REGNO (dest)]) = 0;
       }
       break;
@@ -438,8 +466,13 @@ copy_for_inline (orig)
 	  XEXP (x, i) = copy_for_inline (XEXP (x, i));
 	  break;
 
+	case 'u':
+	  /* Change any references to old-insns to point to the
+	     corresponding copied insns.  */
+	  return insn_map[INSN_UID (XEXP (x, i))];
+
 	case 'E':
-	  if (XVEC (x, i) != NULL)
+	  if (XVEC (x, i) != NULL && XVECLEN (x, i) != 0)
 	    {
 	      register int j;
 
@@ -489,12 +522,15 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
   rtx *arg_vec;
   rtx return_label = 0;
   rtx follows_call = 0;
+  rtx this_struct_value_rtx = 0;
 
   if (max_regno < FIRST_PSEUDO_REGISTER)
-    return (rtx)-1;
+    abort ();
 
   nargs = list_length (DECL_ARGUMENTS (fndecl));
 
+  first_parm_offset = FIRST_PARM_OFFSET (fndecl);
+
   /* We expect PARMS to have the right length; don't crash if not.  */
   if (list_length (parms) != nargs)
     return (rtx)-1;
@@ -517,10 +553,9 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
     {
       tree arg = TREE_VALUE (actual); /* this has already been converted */
       enum machine_mode tmode = TYPE_MODE (TREE_TYPE (formal));
-      tree decl = formal;
       rtx copy;
 
-      emit_note (DECL_SOURCE_FILE (decl), DECL_SOURCE_LINE (decl));
+      emit_note (DECL_SOURCE_FILE (formal), DECL_SOURCE_LINE (formal));
 
       if (TREE_ADDRESSABLE (formal))
 	{
@@ -562,7 +597,17 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 
   /* Pass the function the address in which to return a structure value.  */
   if (structure_value_addr)
-    emit_move_insn (struct_value_rtx, structure_value_addr);
+    {
+      if (GET_CODE (struct_value_rtx) == MEM)
+	{
+	  this_struct_value_rtx = force_reg (Pmode, structure_value_addr);
+	}
+      else
+	{
+	  this_struct_value_rtx = struct_value_rtx;
+	  emit_move_insn (this_struct_value_rtx, structure_value_addr);
+	}
+    }
 
   /* Now prepare for copying the insns.
      Set up reg_map, parm_map and label_map saying how to translate
@@ -576,21 +621,63 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
       tree decl = DECL_ARGUMENTS (fndecl);
       tree last = tree_last (decl);
       int offset = FUNCTION_ARGS_SIZE (header);
+
       parm_map =
 	(rtx *)alloca ((offset / UNITS_PER_WORD) * sizeof (rtx));
       bzero (parm_map, (offset / UNITS_PER_WORD) * sizeof (rtx));
-      parm_map -= FIRST_PARM_OFFSET / UNITS_PER_WORD;
+      parm_map -= first_parm_offset / UNITS_PER_WORD;
 
       for (formal = decl, i = 0; formal; formal = TREE_CHAIN (formal), i++)
 	{
 	  /* Create an entry in PARM_MAP that says what pseudo register
 	     is associated with an address we might compute.  */
-	  parm_map[DECL_OFFSET (formal) / BITS_PER_WORD] = arg_vec[i];
+	  if (DECL_OFFSET (formal) >= 0)
+	    {
+	      /* This parameter has a home in the stack.  */
+	      parm_map[DECL_OFFSET (formal) / BITS_PER_WORD] = arg_vec[i];
+	    }
+	  else
+	    {
+	      /* Parameter that was passed in a register;
+		 does it have a home on the stack (as a local)?  */
+	      rtx frtx = DECL_RTL (formal);
+	      rtx offset = 0;
+	      if (GET_CODE (frtx) == MEM)
+		{
+		  frtx = XEXP (frtx, 0);
+		  if (GET_CODE (frtx) == PLUS)
+		    {
+		      if (XEXP (frtx, 0) == frame_pointer_rtx
+			  && GET_CODE (XEXP (frtx, 1)) == CONST_INT)
+			offset = XEXP (frtx, 1);
+		      else if (XEXP (frtx, 1) == frame_pointer_rtx
+			       && GET_CODE (XEXP (frtx, 0)) == CONST_INT)
+			offset = XEXP (frtx, 0);
+		    }
+		  if (offset)
+		    parm_map[INTVAL (offset) / UNITS_PER_WORD] = arg_vec[i];
+		  else abort ();
+		}
+	      else if (GET_CODE (frtx) != REG)
+		abort ();
+	    }
 	  /* Create an entry in REG_MAP that says what rtx is associated
 	     with a pseudo register from the function being inlined.  */
 	  if (GET_CODE (DECL_RTL (formal)) == REG)
 	    reg_map[REGNO (DECL_RTL (formal))] = arg_vec[i];
 	}
+
+      /* Make certain that we can map struct_value_{incoming_rtx,rtx},
+	 and map it.  If it is a hard register, it is mapped automagically.  */
+      if (GET_CODE (struct_value_incoming_rtx) == REG)
+	;
+      else if (GET_CODE (struct_value_incoming_rtx) == MEM
+	       && XEXP (XEXP (struct_value_incoming_rtx, 0), 0) == frame_pointer_rtx
+	       && GET_CODE (XEXP (XEXP (struct_value_incoming_rtx, 0), 1)) == CONST_INT)
+	parm_map[INTVAL (XEXP (XEXP (struct_value_incoming_rtx, 0), 1)) / UNITS_PER_WORD]
+	  = this_struct_value_rtx;
+      else
+	abort ();
     }
   else
     {
@@ -603,14 +690,40 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
   for (i = min_labelno; i < max_labelno; i++)
     label_map[i] = gen_label_rtx ();
 
+  /* As we copy insns, record the correspondence, so that inter-insn
+     references can be copied into isomorphic structure.  */
+
+  insn_map = (rtx *) alloca (INSN_UID (header) * sizeof (rtx));
+  bzero (insn_map, INSN_UID (header) * sizeof (rtx));
+
   /* Set up a target to translate the inline function's value-register.  */
 
   if (structure_value_addr != 0 || TYPE_MODE (type) == VOIDmode)
     inline_target = 0;
-  else if (target && GET_MODE (target) == TYPE_MODE (type))
-    inline_target = target;
   else
-    inline_target = gen_reg_rtx (TYPE_MODE (type));
+    {
+      /* Machine mode function was declared to return.   */
+      enum machine_mode departing_mode = TYPE_MODE (type);
+      /* (Possibly wider) machine mode it actually computes
+	 (for the sake of callers that fail to declare it right).  */
+      enum machine_mode arriving_mode
+	= TYPE_MODE (TREE_TYPE (DECL_RESULT (fndecl)));
+
+      /* Don't use MEMs as direct targets because on some machines
+	 substituting a MEM for a REG makes invalid insns.
+	 Let the combiner substitute the MEM if that is valid.  */
+      if (target && GET_CODE (target) == REG
+	  && GET_MODE (target) == departing_mode)
+	inline_target = target;
+      else
+	inline_target = target = gen_reg_rtx (departing_mode);
+
+      /* If function's value was promoted before return,
+	 avoid machine mode mismatch when we substitute INLINE_TARGET.
+	 But TARGET is what we will return to the caller.  */
+      if (arriving_mode != departing_mode)
+	inline_target = gen_rtx (SUBREG, arriving_mode, target, 0);
+    }
 
   /* We are about to make space in this function's stack frame
      for a copy of the stack frame of the inline function.
@@ -687,9 +800,10 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 
 	  /* The (USE (REG n)) at return from the function should be ignored
 	     since we are changing (REG n) into inline_target.  */
+	  copy = 0;
 	  if (GET_CODE (pattern) == USE
 	      && GET_CODE (XEXP (pattern, 0)) == REG
-	      && FUNCTION_VALUE_REGNO_P (REGNO (XEXP (pattern, 0))))
+	      && REG_FUNCTION_VALUE_P (XEXP (pattern, 0)))
 	    break;
 
 	  /* Try to do some quick constant folding here.
@@ -724,6 +838,10 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 	  break;
 
 	case CALL_INSN:
+#if 0
+	  /* This should no longer be necessary now that references
+	     to this function's return value are flagged to distinguish
+	     them from other references to the same hard register.  */
 	  {
 	    rtx newbod;
 	    /* If the call's body is (set (reg...) (call...)),
@@ -733,10 +851,28 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 	    if (GET_CODE (PATTERN (insn)) == SET)
 	      newbod = gen_rtx (SET, VOIDmode, SET_DEST (PATTERN (insn)),
 				copy_rtx_and_substitute (SET_SRC (PATTERN (insn))));
+	    else if (GET_CODE (PATTERN (insn)) == PARALLEL
+		     && GET_CODE (XVECEXP (PATTERN (insn), 0, 0)) == SET)
+	      {
+		register int j;
+		rtx newelem;
+		newbod = gen_rtx (PARALLEL, VOIDmode,
+				  rtvec_alloc (XVECLEN (PATTERN (insn), 0)));
+		newelem = gen_rtx (SET, VOIDmode,
+				   SET_DEST (XVECEXP (PATTERN (insn), 0, 0)),
+				   copy_rtx_and_substitute (SET_SRC (XVECEXP (PATTERN (insn), 0, 0))));
+		XVECEXP (newbod, 0, 0) = newelem;
+		for (j = 1; j < XVECLEN (newbod, 0); j++)
+		  XVECEXP (newbod, 0, j)
+		    = copy_rtx_and_substitute (XVECEXP (PATTERN (insn), 0, j));
+	      }
 	    else
 	      newbod = copy_rtx_and_substitute (PATTERN (insn));
 	    copy = emit_call_insn (newbod);
 	  }
+#else /* 1 */
+	  copy = emit_call_insn (copy_rtx_and_substitute (PATTERN (insn)));
+#endif /* 1 */
 	  copy->integrated = 1;
 	  /* Special handling needed for the following INSN depending on
 	     whether it copies the value from the fcn return reg.  */
@@ -745,22 +881,27 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 	  break;
 
 	case CODE_LABEL:
-	  emit_label (label_map[CODE_LABEL_NUMBER (insn)]);
+	  copy = emit_label (label_map[CODE_LABEL_NUMBER (insn)]);
 	  follows_call = 0;
 	  break;
 
 	case BARRIER:
-	  emit_barrier ();
+	  copy = emit_barrier ();
 	  break;
 
 	case NOTE:
-	  emit_note (NOTE_SOURCE_FILE (insn), NOTE_LINE_NUMBER (insn));
+	  if (NOTE_LINE_NUMBER (insn) != NOTE_INSN_FUNCTION_END)
+	    copy = emit_note (NOTE_SOURCE_FILE (insn), NOTE_LINE_NUMBER (insn));
+	  else
+	    copy = 0;
 	  break;
 
 	default:
 	  abort ();
 	  break;
 	}
+
+      insn_map[INSN_UID (insn)] = copy;
     }
 
   if (return_label)
@@ -789,7 +930,7 @@ expand_inline_function (fndecl, parms, target, ignore, type, structure_value_add
 		      memory_address (BLKmode, structure_value_addr));
     }
 
-  return inline_target;
+  return target;
 }
 
 /* Given a chain of PARM_DECLs, ARGS, and a vector of RTL homes VEC,
@@ -881,7 +1022,7 @@ copy_rtx_and_substitute (orig)
       regno = REGNO (orig);
       if (regno < FIRST_PSEUDO_REGISTER)
 	{
-	  if (FUNCTION_VALUE_REGNO_P (regno))
+	  if (REG_FUNCTION_VALUE_P (orig))
 	    return inline_target;
 	  if (regno == FRAME_POINTER_REGNUM)
 	    return plus_constant (orig, fp_delta);
@@ -907,6 +1048,20 @@ copy_rtx_and_substitute (orig)
     case SYMBOL_REF:
       return orig;
 
+    case CALL:
+      /* This is given special treatment because the first
+	 operand of a CALL is a (MEM ...) which may get
+	 forced into a register for cse.  This is undesirable
+	 if function-address cse isn't wanted or if we won't do cse.  */
+#ifndef NO_FUNCTION_CSE
+      if (! (optimize && ! flag_no_function_cse))
+#endif
+	return gen_rtx (CALL, GET_MODE (orig),
+			gen_rtx (MEM, GET_MODE (XEXP (orig, 0)),
+				 copy_rtx_and_substitute (XEXP (XEXP (orig, 0), 0))),
+			copy_rtx_and_substitute (XEXP (orig, 1)));
+      break;
+
     case PLUS:
       /* Note:  the PLUS case is not nearly as careful as the MEM
 	 case in terms of preserving addresses.  The reason for this
@@ -943,7 +1098,7 @@ copy_rtx_and_substitute (orig)
 				       c + fp_delta));
 	    }
 	  copy = copy_rtx_and_substitute (copy);
-	  temp = gen_rtx (PLUS, mode, frame_pointer_rtx, copy);
+	  temp = force_reg (mode, gen_rtx (PLUS, mode, frame_pointer_rtx, copy));
 	  return plus_constant (temp, fp_delta);
 	}
       else if (reg_mentioned_p (frame_pointer_rtx, orig)
@@ -984,13 +1139,30 @@ copy_rtx_and_substitute (orig)
 			copy_rtx_and_substitute (XEXP (orig, 1)));
 
       return temp;
-      
+
     case MEM:
       /* Take care of easiest case here.  */
       copy = XEXP (orig, 0);
       if (copy == frame_pointer_rtx || copy == arg_pointer_rtx)
 	return gen_rtx (MEM, mode,
 			plus_constant (frame_pointer_rtx, fp_delta));
+
+      /* Allow a pushing-address even if that is not valid as an
+	 ordinary memory address.  It indicates we are inlining a special
+	 push-insn.  */
+#ifdef STACK_GROWS_DOWNWARD
+      if (GET_CODE (copy) == PRE_DEC && XEXP (copy, 0) == stack_pointer_rtx)
+	return orig;
+#else
+      if (GET_CODE (copy) == PRE_INC && XEXP (copy, 0) == stack_pointer_rtx)
+	return orig;
+#endif
+
+      /* If this is some other sort of address that isn't generally valid,
+	 break out all the registers referred to.  */
+      if (! memory_address_p (mode, copy))
+	return gen_rtx (MEM, mode, copy_address (copy));
+
       if (GET_CODE (copy) == PLUS)
 	{
 	  if (XEXP (copy, 0) == frame_pointer_rtx
@@ -1010,8 +1182,7 @@ copy_rtx_and_substitute (orig)
 		{
 		  int c = INTVAL (copy);
 
-		  if (reg == arg_pointer_rtx
-		      && c >= FIRST_PARM_OFFSET)
+		  if (reg == arg_pointer_rtx && c >= first_parm_offset)
 		    {
 		      copy = parm_map[c / UNITS_PER_WORD];
 
@@ -1095,7 +1266,6 @@ copy_rtx_and_substitute (orig)
 
       switch (*format_ptr++)
 	{
-	case 'u':
 	case '0':
 	  break;
 
@@ -1103,9 +1273,106 @@ copy_rtx_and_substitute (orig)
 	  XEXP (copy, i) = copy_rtx_and_substitute (XEXP (orig, i));
 	  break;
 
+	case 'u':
+	  /* Change any references to old-insns to point to the
+	     corresponding copied insns.  */
+	  return insn_map[INSN_UID (XEXP (orig, i))];
+
 	case 'E':
 	  XVEC (copy, i) = XVEC (orig, i);
-	  if (XVEC (orig, i) != NULL)
+	  if (XVEC (orig, i) != NULL && XVECLEN (orig, i) != 0)
+	    {
+	      XVEC (copy, i) = rtvec_alloc (XVECLEN (orig, i));
+	      for (j = 0; j < XVECLEN (copy, i); j++)
+		XVECEXP (copy, i, j) = copy_rtx_and_substitute (XVECEXP (orig, i, j));
+	    }
+	  break;
+
+	case 'i':
+	  XINT (copy, i) = XINT (orig, i);
+	  break;
+
+	case 's':
+	  XSTR (copy, i) = XSTR (orig, i);
+	  break;
+
+	default:
+	  abort ();
+	}
+    }
+  return copy;
+}
+
+/* Like copy_rtx_and_substitute but produces different output, suitable
+   for an ideosyncractic address that isn't memory_address_p.
+   The output resembles the input except that REGs and MEMs are replaced
+   with new psuedo registers.  All the "real work" is done in separate
+   insns which set up the values of these new registers.  */
+
+static rtx
+copy_address (orig)
+     register rtx orig;
+{
+  register rtx copy, temp;
+  register int i, j;
+  register RTX_CODE code;
+  register enum machine_mode mode;
+  register char *format_ptr;
+  int regno;
+
+  if (orig == 0)
+    return 0;
+
+  code = GET_CODE (orig);
+  mode = GET_MODE (orig);
+
+  switch (code)
+    {
+    case REG:
+    case MEM:
+      return copy_to_reg (copy_rtx_and_substitute (code));
+
+    case CODE_LABEL:
+    case LABEL_REF:
+      return copy_rtx_and_substitute (code);
+
+    case PC:
+    case CC0:
+    case CONST_INT:
+    case CONST_DOUBLE:
+    case SYMBOL_REF:
+      return orig;
+    }
+
+  copy = rtx_alloc (code);
+  PUT_MODE (copy, mode);
+  copy->in_struct = orig->in_struct;
+  copy->volatil = orig->volatil;
+  copy->unchanging = orig->unchanging;
+
+  format_ptr = GET_RTX_FORMAT (GET_CODE (copy));
+
+  for (i = 0; i < GET_RTX_LENGTH (GET_CODE (copy)); i++)
+    {
+      rtx new;
+
+      switch (*format_ptr++)
+	{
+	case '0':
+	  break;
+
+	case 'e':
+	  XEXP (copy, i) = copy_rtx_and_substitute (XEXP (orig, i));
+	  break;
+
+	case 'u':
+	  /* Change any references to old-insns to point to the
+	     corresponding copied insns.  */
+	  return insn_map[INSN_UID (XEXP (orig, i))];
+
+	case 'E':
+	  XVEC (copy, i) = XVEC (orig, i);
+	  if (XVEC (orig, i) != NULL && XVECLEN (orig, i) != 0)
 	    {
 	      XVEC (copy, i) = rtvec_alloc (XVECLEN (orig, i));
 	      for (j = 0; j < XVECLEN (copy, i); j++)
@@ -1122,9 +1389,6 @@ copy_rtx_and_substitute (orig)
 	  break;
 
 	default:
-	  fprintf (stderr,
-		   "switch format wrong in rtl2.copy_rtx_and_substitute(). format was: %c.\n",
-		   format_ptr[-1]);
 	  abort ();
 	}
     }
@@ -1366,6 +1630,8 @@ output_inline_function (fndecl)
 
   temporary_allocation ();
 
+  current_function_decl = fndecl;
+
   /* This call is only used to initialize global variables.
      The rtl code it emits will be discarded below.  */
   expand_function_start (fndecl);
@@ -1385,5 +1651,7 @@ output_inline_function (fndecl)
   /* Compile this function all the way down to assembly code.  */
   rest_of_compilation (fndecl);
 
+  current_function_decl = 0;
+
   permanent_allocation ();
 }
