@@ -135,6 +135,7 @@ static FILE *loop_dump_stream;
 
 static rtx verify_loop ();
 static int invariant_p ();
+static int consec_sets_invariant_p ();
 static int can_jump_into_range_p ();
 static void count_loop_regs_set ();
 static void note_addr_stored ();
@@ -210,10 +211,18 @@ loop_optimize (f, nregs, dumpfile)
 
   for (insn = last_insn; insn; insn = PREV_INSN (insn))
     if (GET_CODE (insn) == NOTE
-	&& NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_BEG
+	&& NOTE_LINE_NUMBER (insn) == NOTE_INSN_LOOP_BEG)
+      {
 	/* Make sure it really is a loop -- no jumps in from outside.  */
-	&& (end = verify_loop (f, insn)))
-      scan_loop (insn, end, nregs);
+	end = verify_loop (f, insn);
+	if (end != 0)
+	  /* If so, optimize this loop.  */
+	  scan_loop (insn, end, nregs);
+	else if (loop_dump_stream)
+	  fprintf (loop_dump_stream,
+		   "\nLoop at %d ignored due to multiple entry points.\n",
+		   INSN_UID (insn));
+      }
 }
 
 /* Optimize one loop whose start is LOOP_START and end is END.
@@ -229,6 +238,9 @@ scan_loop (loop_start, end, nregs)
   register rtx p = NEXT_INSN (loop_start);
   /* 1 if we are scanning insns that could be executed zero times.  */
   int maybe_never = 0;
+  /* 1 if we are scanning insns that might never be executed
+     due to a subroutine call which might exit before they are reached.  */
+  int call_passed = 0;
   /* For a rotated loop that is entered near the bottom,
      this is the label at the top.  Otherwise it is zero.  */
   rtx loop_top = 0;
@@ -248,7 +260,7 @@ scan_loop (loop_start, end, nregs)
      0 indicates an invariant register; -1 a conditionally invariant one.  */
   short *n_times_set;
   /* Indexed by register number, contains the number of times the reg
-     was set during the loop being scanned, not counting changes due
+     was used during the loop being scanned, not counting changes due
      to moving these insns out of the loop.  */
   short *n_times_used;
   /* Indexed by register number, contains 1 for a register whose
@@ -263,6 +275,9 @@ scan_loop (loop_start, end, nregs)
      since in that case saving an insn makes more difference
      and more registers are available.  */
   int threshold = loop_has_call ? 17 : 34;
+  /* Nonzero if the insn that jumps into the real loop
+     is not the very first thing after the loop-beginning note.  */
+  int something_before_entry_jump = 0;
 
   n_times_set = (short *) alloca (nregs * sizeof (short));
   n_times_used = (short *) alloca (nregs * sizeof (short));
@@ -272,7 +287,11 @@ scan_loop (loop_start, end, nregs)
      to a test at the end.  */
   while (p != end
 	 && GET_CODE (p) != CODE_LABEL && GET_CODE (p) != JUMP_INSN)
-    p = NEXT_INSN (p);
+    {
+      if (GET_CODE (p) == CALL_INSN || GET_CODE (p) == INSN)
+	something_before_entry_jump = 1;
+      p = NEXT_INSN (p);
+    }
 
   /* "Loop" contains neither jumps nor labels;
      it must have been a dummy.  Think no more about it.  */
@@ -294,7 +313,8 @@ scan_loop (loop_start, end, nregs)
 	 If we see one, this must not be a real loop.  */
       if (GET_CODE (loop_top) != BARRIER)
 	return;
-      p = JUMP_LABEL (p);
+      /* Get the label at which the loop is entered.  */
+      p = XEXP (SET_SRC (PATTERN (p)), 0);
       /* Check to see whether the jump actually
 	 jumps out of the loop (meaning it's no loop).
 	 This case can happen for things like
@@ -321,7 +341,8 @@ scan_loop (loop_start, end, nregs)
 	 with a new test to jump around it entirely.
 	 (The latter is considered outside the loop.)
 	 If this is done, we no longer enter with a jump.  */
-      if (loop_skip_over (loop_start, end, loop_entry_jump))
+      if (! something_before_entry_jump
+	  && loop_skip_over (loop_start, end, loop_entry_jump))
 	loop_top = 0;
       else
 	/* We really do enter with a jump;
@@ -382,6 +403,7 @@ scan_loop (loop_start, end, nregs)
 	  && GET_CODE (SET_DEST (PATTERN (p))) == REG
 	  && ! may_not_move[REGNO (SET_DEST (PATTERN (p)))])
 	{
+	  int tem1 = 0;
 	  /* If this register is used or set outside the loop,
 	     then we can move it only if we know this insn is
 	     executed exactly once per iteration,
@@ -393,14 +415,19 @@ scan_loop (loop_start, end, nregs)
 	      && (maybe_never
 		  || loop_reg_used_before_p (p, loop_start, scan_start, end)))
 	    ;
-	  /* If this can cause a trap (such as divide by zero), can't move it
-	     unless it's guaranteed to be executed once loop is entered.  */
-	  else if (maybe_never && may_trap_p (SET_SRC (PATTERN (p))))
-	    ;
 	  else if ((tem = invariant_p (SET_SRC (PATTERN (p)), n_times_set))
 		   && (n_times_set[REGNO (SET_DEST (PATTERN (p)))] == 1
-		       || consec_sets_invariant_p (SET_DEST (PATTERN (p)),
-						   p, n_times_set)))
+		       || (tem1
+			   = consec_sets_invariant_p (SET_DEST (PATTERN (p)),
+						      n_times_set[REGNO (SET_DEST (PATTERN (p)))],
+						      p, n_times_set)))
+		   /* If the insn can cause a trap (such as divide by zero),
+		      can't move it unless it's guaranteed to be executed
+		      once loop is entered.  Even a function call might
+		      prevent the trap insn from being reached
+		      (since it might exit!)  */
+		   && ! ((maybe_never || call_passed)
+			 && may_trap_p (SET_SRC (PATTERN (p)))))
 	    {
 	      register struct movable *m;
 	      register int regno = REGNO (SET_DEST (PATTERN (p)));
@@ -414,7 +441,9 @@ scan_loop (loop_start, end, nregs)
 	      m->forces = 0;
 	      m->partial = 0;
 	      m->regno = regno;
-	      m->cond = (tem > 1);
+	      /* Set M->cond if either invariant_p or consec_sets_invariant_p
+		 returned 2 (only conditionally invariant).  */
+	      m->cond = ((tem | tem1) > 1);
 	      m->global = (uid_luid[regno_last_uid[regno]] > INSN_LUID (end)
 			   || uid_luid[regno_first_uid[regno]] < INSN_LUID (loop_start));
 	      m->match = 0;
@@ -494,6 +523,10 @@ scan_loop (loop_start, end, nregs)
 		}
 	    }
 	}
+      /* Past a call insn, we get to insns which might not be executed
+	 because the call might exit.  This matters for insns that trap.  */
+      else if (GET_CODE (p) == CALL_INSN)
+	call_passed = 1;
       /* Past a label or a jump, we get to insns for which we
 	 can't count on whether or how many times they will be
 	 executed during each iteration.  Therefore, we can
@@ -646,7 +679,11 @@ move_movables (movables, n_times_set, n_times_used, threshold,
 
       if (!m->done
 	  && (! m->cond
-	      || 1 == invariant_p (SET_SRC (PATTERN (m->insn)), n_times_set))
+	      || (1 == invariant_p (SET_SRC (PATTERN (m->insn)), n_times_set)
+		  && (m->consec == 0
+		      || 1 == consec_sets_invariant_p (SET_DEST (PATTERN (m->insn)),
+						       m->consec + 1,
+						       m->insn, n_times_set))))
 	  && (! m->forces || m->forces->done))
 	{
 	  register int regno;
@@ -862,6 +899,7 @@ replace_regs (x, reg_map)
   return x;
 }
 
+#if 0
 /* P is an instruction that sets a register to the result of a ZERO_EXTEND.
    Replace it with an instruction to load just the low bytes
    if the machine supports such an instruction,
@@ -899,6 +937,7 @@ constant_high_bytes (p, loop_start)
       PATTERN (p) = new;
     }
 }
+#endif
 
 /* Verify that the ostensible loop starting at START
    really is a loop: nothing jumps into it from outside.
@@ -1110,8 +1149,9 @@ invariant_p (x, n_times_set)
       return 0;
 
     case REG:
-      if (x == frame_pointer_rtx || x == arg_pointer_rtx
-	  || x->unchanging)
+      /* We used to check x->unchanging here, but that is invalid
+	 since the reg might be set by initialization within the loop.  */
+      if (x == frame_pointer_rtx || x == arg_pointer_rtx)
 	return 1;
       if (n_times_set[REGNO (x)] == -1)
 	return 2;
@@ -1125,10 +1165,12 @@ invariant_p (x, n_times_set)
       /* Don't mess with volatile memory references.  */
       if (x->volatil)
 	return 0;
+#if 0
       /* If it's declared read-only, it is invariant
 	 if its address is invariant.  */
       if (x->unchanging)
 	return invariant_p (XEXP (x, 0), n_times_set);
+#endif
       /* A store in a varying-address aggregate component
 	 could clobber anything except a scalar with a fixed address.  */
       if (unknown_aggregate_altered
@@ -1213,24 +1255,28 @@ addr_overlap_p (other, base, size)
   return refers_to_mem_p (other, base, start, end);
 }
 
-/* Return 1 if all the insns in the loop that set REG
+/* Return nonzero if all the insns in the loop that set REG
    are INSN and the immediately following insns,
    and if each of those insns sets REG in an invariant way
    according to TABLE (not counting uses of REG in them).
 
+   The value is 2 if some of these insns are only conditionally invariant.
+
    We assume that INSN itself is the first set of REG
    and that its source is invariant.  */
 
 static int
-consec_sets_invariant_p (reg, insn, table)
+consec_sets_invariant_p (reg, n_sets, insn, table)
+     int n_sets;
      rtx reg, insn;
      short *table;
 {
   register rtx p = insn;
   register int regno = REGNO (reg);
   /* Number of sets we have to insist on finding after INSN.  */
-  int count = table[regno] - 1;
+  int count = n_sets - 1;
   int old = table[regno];
+  int tem = 0;
 
   table[regno] = 0;
 
@@ -1242,7 +1288,7 @@ consec_sets_invariant_p (reg, insn, table)
       if (code == INSN && GET_CODE (PATTERN (p)) == SET
 	  && GET_CODE (SET_DEST (PATTERN (p))) == REG
 	  && REGNO (SET_DEST (PATTERN (p))) == regno
-	  && invariant_p (SET_SRC (PATTERN (p)), table))
+	  && (tem |= invariant_p (SET_SRC (PATTERN (p)), table)))
 	count--;
       else if (code != NOTE)
 	{
@@ -1252,7 +1298,8 @@ consec_sets_invariant_p (reg, insn, table)
     }
 
   table[regno] = old;
-  return 1;
+  /* If invariant_p ever returned 2, we return 2.  */
+  return 1 + (tem & 2);
 }
 
 #if 0
@@ -1432,14 +1479,34 @@ loop_reg_used_before_p (insn, loop_start, scan_start, loop_end)
     return reg_used_between_p (reg, scan_start, insn);
 }
 
-/* Return nonzero if X is a division that might trap.  */
+/* Return nonzero if evaluating rtx X might cause a trap.  */
 
 static int
 may_trap_p (x)
      rtx x;
 {
-  switch (GET_CODE (x))
+  int i;
+  enum rtx_code code = GET_CODE (x);
+  char *fmt;
+
+  switch (code)
     {
+      /* Handle these cases fast.  */
+    case CONST_INT:
+    case CONST_DOUBLE:
+    case SYMBOL_REF:
+    case LABEL_REF:
+    case CONST:
+    case PC:
+    case CC0:
+    case REG:
+      return 0;
+
+      /* Memory ref can trap unless it's a static var or a stack slot.  */
+    case MEM:
+      return rtx_varies_p (XEXP (x, 0));
+
+      /* Division by a non-constant might trap.  */
     case DIV:
     case MOD:
     case UDIV:
@@ -1447,6 +1514,27 @@ may_trap_p (x)
       if (! CONSTANT_P (XEXP (x, 1))
 	  && GET_CODE (XEXP (x, 1)) != CONST_DOUBLE)
 	return 1;
+    default:
+      /* Any floating arithmetic may trap.  */
+      if (GET_MODE_CLASS (GET_MODE (x)) == MODE_FLOAT)
+	return 1;
+    }
+
+  fmt = GET_RTX_FORMAT (code);
+  for (i = GET_RTX_LENGTH (code) - 1; i >= 0; i--)
+    {
+      if (fmt[i] == 'e')
+	{
+	  if (may_trap_p (XEXP (x, i)))
+	    return 1;
+	}
+      else if (fmt[i] == 'E')
+	{
+	  register int j;
+	  for (j = 0; j < XVECLEN (x, i); j++)
+	    if (may_trap_p (XVECEXP (x, i, j)))
+	      return 1;
+	}
     }
   return 0;
 }
