@@ -252,6 +252,17 @@ immediate_operand (op, mode)
 	  && LEGITIMATE_CONSTANT_P (op));
 }
 
+/* Return 1 if OP is a general operand that is not an immediate operand.  */
+
+int
+nonimmediate_operand (op, mode)
+     register rtx op;
+     enum machine_mode mode;
+{
+  return (general_operand (op, mode)
+	  && ! CONSTANT_P (op) && GET_CODE (op) != CONST_DOUBLE);
+}
+
 /* Return 1 if OP is a register reference or immediate value of mode MODE.  */
 
 int
@@ -299,7 +310,7 @@ push_operand (op, mode)
   if (GET_CODE (op) != PRE_INC)
     return 0;
 #endif
-  return REGNO (XEXP (op, 0)) == STACK_POINTER_REGNUM;
+  return XEXP (op, 0) == stack_pointer_rtx;
 }
 
 /* Return 1 if ADDR is a valid memory address for mode MODE.  */
@@ -327,7 +338,31 @@ memory_operand (op, mode)
      register rtx op;
      enum machine_mode mode;
 {
-  return GET_CODE (op) == MEM && general_operand (op, mode);
+  enum rtx_code code = GET_CODE (op);
+  int mode_altering_drug = 0;
+
+  while (code == SUBREG)
+    {
+      op = SUBREG_REG (op);
+      code = GET_CODE (op);
+      mode_altering_drug = 1;
+    }
+
+  return (GET_CODE (op) == MEM && general_operand (op, mode)
+	  && ! (mode_altering_drug
+		&& mode_dependent_address_p (XEXP (op, 0))));
+}
+
+/* Return 1 if OP is a valid indirect memory reference with mode MODE;
+   that is, a memory reference whose address is a general_operand.  */
+
+int
+indirect_operand (op, mode)
+     register rtx op;
+     enum machine_mode mode;
+{
+  return (GET_CODE (op) == MEM && GET_MODE (op) == mode
+	  && general_operand (XEXP (op, 0), Pmode));
 }
 
 /* If BODY is an insn body that uses ASM_OPERANDS,
@@ -340,15 +375,26 @@ asm_noperands (body)
 {
   int noperands;
 
+  if (GET_CODE (body) == ASM_OPERANDS)
+    /* No output operands: return number of input operands.  */
+    return XVECLEN (body, 3);
   if (GET_CODE (body) == SET && GET_CODE (SET_SRC (body)) == ASM_OPERANDS)
     /* Single output operand: BODY is (set OUTPUT (asm_operands ...)).  */
     return XVECLEN (SET_SRC (body), 3) + 1;
   else if (GET_CODE (body) == PARALLEL
 	   && GET_CODE (XVECEXP (body, 0, 0)) == SET
 	   && GET_CODE (SET_SRC (XVECEXP (body, 0, 0))) == ASM_OPERANDS)
-    /* Multiple outputs: BODY is
-       (parallel [(set OUTPUT0 (asm_operands ...)) ...]).  */
-    return XVECLEN (SET_SRC (XVECEXP (body, 0, 0)), 3) + XVECLEN (body, 0);
+    {
+      /* Multiple output operands:
+	 body is [(set OUTPUT (asm_operands ...))... (clobber (reg ...))...].
+	 Count backwards through CLOBBERs to determine number of SETs.  */
+      int i;
+      for (i = XVECLEN (body, 0) - 1; i >= 0; i--)
+	if (GET_CODE (XVECEXP (body, 0, i)) == SET)
+	  break;
+      
+      return XVECLEN (SET_SRC (XVECEXP (body, 0, 0)), 3) + i + 1;
+    }
   else
     return 0;
 }
@@ -408,13 +454,52 @@ decode_asm_operands (body, operands, operand_locs, constraints, modes)
 	modes[0] = GET_MODE (SET_DEST (body));
       template = XSTR (asmop, 0);
     }
+  else if (GET_CODE (body) == ASM_OPERANDS)
+    {
+      rtx asmop = body;
+      /* No output operands: BODY is (asm_operands ....).  */
+
+      noperands = XVECLEN (asmop, 3);
+
+      /* The input operands are found in the 1st element vector.  */
+      /* Constraints for inputs are in the 2nd element vector.  */
+      for (i = 0; i < noperands; i++)
+	{
+	  if (operand_locs)
+	    operand_locs[i] = &XVECEXP (asmop, 3, i);
+	  if (operands)
+	    operands[i] = XVECEXP (asmop, 3, i);
+	  if (constraints)
+	    constraints[i] = XSTR (XVECEXP (asmop, 4, i), 0);
+	  if (modes)
+	    modes[i] = GET_MODE (XVECEXP (asmop, 4, i));
+	}
+      template = XSTR (asmop, 0);
+    }
   else
     {
       rtx asmop = SET_SRC (XVECEXP (body, 0, 0));
-      int nout = XVECLEN (body, 0);
+      int nparallel = XVECLEN (body, 0); /* Includes CLOBBERs.  */
       int nin = XVECLEN (asmop, 3);
+      int nout = 0;		/* Does not include CLOBBERs.  */
 
-      noperands = XVECLEN (asmop, 3) + XVECLEN (body, 0);
+      /* The outputs are in the SETs.
+	 Their constraints are in the ASM_OPERANDS itself.  */
+      for (i = 0; i < nparallel; i++)
+	{
+	  if (GET_CODE (XVECEXP (body, 0, i)) == CLOBBER)
+	    break;		/* Past last SET */
+	  
+	  if (operands)
+	    operands[i] = SET_DEST (XVECEXP (body, 0, i));
+	  if (operand_locs)
+	    operand_locs[i] = &SET_DEST (XVECEXP (body, 0, i));
+	  if (constraints)
+	    constraints[i] = XSTR (SET_SRC (XVECEXP (body, 0, i)), 1);
+	  if (modes)
+	    modes[i] = GET_MODE (SET_DEST (XVECEXP (body, 0, i)));
+	  nout++;
+	}
 
       /* The input operands are found in the 1st element vector.  */
       /* Constraints for inputs are in the 2nd element vector.  */
@@ -429,19 +514,7 @@ decode_asm_operands (body, operands, operand_locs, constraints, modes)
 	  if (modes)
 	    modes[i + nout] = GET_MODE (XVECEXP (asmop, 4, i));
 	}
-      /* The outputs are in the SETs.
-	 Their constraints are in the ASM_OPERANDS itself.  */
-      for (i = 0; i < nout; i++)
-	{
-	  if (operands)
-	    operands[i] = SET_DEST (XVECEXP (body, 0, i));
-	  if (operand_locs)
-	    operand_locs[i] = &SET_DEST (XVECEXP (body, 0, i));
-	  if (constraints)
-	    constraints[i] = XSTR (SET_SRC (XVECEXP (body, 0, i)), 1);
-	  if (modes)
-	    modes[i] = GET_MODE (SET_DEST (XVECEXP (body, 0, i)));
-	}
+
       template = XSTR (asmop, 0);
     }
 
