@@ -1,5 +1,5 @@
 /* Compute register class preferences for pseudo-registers.
-   Copyright (C) 1987 Free Software Foundation, Inc.
+   Copyright (C) 1987, 1988 Free Software Foundation, Inc.
 
 This file is part of GNU CC.
 
@@ -66,11 +66,20 @@ HARD_REG_SET call_used_reg_set;
 /* Data for initializing the above.  */
 
 static char initial_call_used_regs[] = CALL_USED_REGISTERS;
+  
+/* Table of register numbers in the order in which to try to use them.  */
+#ifdef REG_ALLOC_ORDER
+int reg_alloc_order[FIRST_PSEUDO_REGISTER] = REG_ALLOC_ORDER;
+#endif
 
 /* For each reg class, a HARD_REG_SET saying which registers are in it.  */
 
 HARD_REG_SET reg_class_contents[] = REG_CLASS_CONTENTS;
 
+/* For each reg class, number of regs it contains.  */
+
+int reg_class_size[N_REG_CLASSES];
+
 /* For each reg class, table listing all the containing classes.  */
 
 enum reg_class reg_class_superclasses[N_REG_CLASSES][N_REG_CLASSES];
@@ -84,8 +93,13 @@ enum reg_class reg_class_subclasses[N_REG_CLASSES][N_REG_CLASSES];
 
 enum reg_class reg_class_subunion[N_REG_CLASSES][N_REG_CLASSES];
 
+/* Array containing all of the register names */
 
-/* Function called only once to initialize the tables above.  */
+char *reg_names[] = REGISTER_NAMES;
+
+
+/* Function called only once to initialize the above data on reg usage.
+   Once this is done, various switches may override.  */
 
 void
 init_reg_sets ()
@@ -95,25 +109,13 @@ init_reg_sets ()
   bcopy (initial_fixed_regs, fixed_regs, sizeof fixed_regs);
   bcopy (initial_call_used_regs, call_used_regs, sizeof call_used_regs);
 
-  /* This macro allows the fixed or call-used registers
-     to depend on target flags.  */
+  /* Compute number of hard regs in each class.  */
 
-#ifdef CONDITIONAL_REGISTER_USAGE
-  CONDITIONAL_REGISTER_USAGE
-#endif
-
-  /* Initialize "constant" tables.  */
-
-  CLEAR_HARD_REG_SET (fixed_reg_set);
-  CLEAR_HARD_REG_SET (call_used_reg_set);
-
-  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
-    {
-      if (fixed_regs[i])
-	SET_HARD_REG_BIT (fixed_reg_set, i);
-      if (call_used_regs[i])
-	SET_HARD_REG_BIT (call_used_reg_set, i);
-    }
+  bzero (reg_class_size, sizeof reg_class_size);
+  for (i = 0; i < N_REG_CLASSES; i++)
+    for (j = 0; j < FIRST_PSEUDO_REGISTER; j++)
+      if (TEST_HARD_REG_BIT (reg_class_contents[i], j))
+	reg_class_size[i]++;
 
   /* Initialize the table of subunions.
      reg_class_subunion[I][J] gets the largest-numbered reg-class
@@ -179,6 +181,36 @@ init_reg_sets ()
 	  *p = (enum reg_class) i;
 	}
     }
+
+}
+
+/* After switches have been processed, which perhaps alter
+   `fixed_regs' and `call_used_regs', convert them to HARD_REG_SETs.  */
+
+void
+init_reg_sets_1 ()
+{
+  register int i;
+
+  /* This macro allows the fixed or call-used registers
+     to depend on target flags.  */
+
+#ifdef CONDITIONAL_REGISTER_USAGE
+  CONDITIONAL_REGISTER_USAGE;
+#endif
+
+  /* Initialize "constant" tables.  */
+
+  CLEAR_HARD_REG_SET (fixed_reg_set);
+  CLEAR_HARD_REG_SET (call_used_reg_set);
+
+  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
+    {
+      if (fixed_regs[i])
+	SET_HARD_REG_BIT (fixed_reg_set, i);
+      if (call_used_regs[i])
+	SET_HARD_REG_BIT (call_used_reg_set, i);
+    }
 }
 
 /* Specify the usage characteristics of the register named NAME.
@@ -190,7 +222,6 @@ fix_register (name, fixed, call_used)
      char *name;
      int fixed, call_used;
 {
-  static char *reg_names[] = REGISTER_NAMES;
   int i;
 
   /* Decode the name and update the primary form of
@@ -209,19 +240,6 @@ fix_register (name, fixed, call_used)
       warning ("unknown register name: %s", name);
       return;
     }
-
-  /* Reinitialize the HARD_REG_SETs that have the same info.  */
-
-  CLEAR_HARD_REG_SET (fixed_reg_set);
-  CLEAR_HARD_REG_SET (call_used_reg_set);
-
-  for (i = 0; i < FIRST_PSEUDO_REGISTER; i++)
-    {
-      if (fixed_regs[i])
-	SET_HARD_REG_BIT (fixed_reg_set, i);
-      if (call_used_regs[i])
-	SET_HARD_REG_BIT (call_used_reg_set, i);
-    }
 }
 
 /* Now the data and code for the `regclass' pass, which happens
@@ -235,6 +253,7 @@ struct savings
 {
   short savings[N_REG_CLASSES];
   short memcost;
+  short nrefs;
 };
 
 static struct savings *savings;
@@ -321,6 +340,8 @@ regclass (f, nregs)
 	if (GET_CODE (insn) == INSN && asm_noperands (PATTERN (insn)) > 0)
 	  {
 	    int noperands = asm_noperands (PATTERN (insn));
+	    /* We don't use alloca because alloca would not free
+	       any of the space until this function returns.  */
 	    rtx *operands = (rtx *) oballoc (noperands * sizeof (rtx));
 	    char **constraints
 	      = (char **) oballoc (noperands * sizeof (char *));
@@ -328,8 +349,7 @@ regclass (f, nregs)
 	    decode_asm_operands (PATTERN (insn), operands, 0, constraints, 0);
 
 	    for (i = noperands - 1; i >= 0; i--)
-	      reg_class_record (operands[i],
-				&constraints[i]);
+	      reg_class_record (operands[i], i, constraints);
 
 	    obfree (operands);
 	  }
@@ -340,8 +360,8 @@ regclass (f, nregs)
 	    insn_extract (insn);
 
 	    for (i = insn_n_operands[insn_code_number] - 1; i >= 0; i--)
-	      reg_class_record (recog_operand[i],
-				&insn_operand_constraint[insn_code_number][i]);
+	      reg_class_record (recog_operand[i], i,
+				insn_operand_constraint[insn_code_number]);
 
 	    /* Improve handling of two-address insns such as
 	       (set X (ashift CONST Y)) where CONST must be made to match X.
@@ -422,44 +442,47 @@ regclass (f, nregs)
       /* Note that best_savings is twice number of places something
 	 is saved.  */
       if ((best_savings - p->savings[(int) GENERAL_REGS]) * 5 < reg_n_refs[i])
-	prefclass[i] = (char) GENERAL_REGS;
+	prefclass[i] = (int) GENERAL_REGS;
       else
-	prefclass[i] = (char) best;
+	prefclass[i] = (int) best;
 #else
-      prefclass[i] = (char) best;
+      /* We cast to (int) because (char) hits bugs in some compilers.  */
+      prefclass[i] = (int) best;
 #endif
 
       /* reg_n_refs + p->memcost measures the cost of putting in memory.
-	 If a GENERAL_REG is no better, don't even try for one.  */
+	 If a GENERAL_REG is no better, don't even try for one.
+	 Since savings and memcost are 2 * number of refs,
+	 this effectively counts each memory operand not needing reloading
+	 as costing 1/2 of a reload insn.  */
       if (reg_n_refs != 0)
 	preferred_or_nothing[i]
-	  = ((best_savings - p->savings[(int) GENERAL_REGS]) / 2 
-	     >= reg_n_refs[i] + p->memcost);
+	  = ((best_savings - p->savings[(int) GENERAL_REGS])
+	     >= p->nrefs + p->memcost);
     }
 #endif /* REGISTER_CONSTRAINTS */
 }
 
 #ifdef REGISTER_CONSTRAINTS
 
-/* Scan an operand OP to which the constraint *CONSTRAINT_LOC should apply
-   and record the preferred register classes from the constraint for OP
-   if OP is a register.  If OP is a memory reference, record suitable
-   preferences for registers used in the address.
+/* Scan an operand OP for register class preferences.
+   OPNO is the operand number, and CONSTRAINTS is the constraint
+   vector for the insn.
 
-   We can deduce both the insn code number and which operand in the insn
-   this is supposed to be from the position of CONSTRAINT_LOC
-   in the table of constraints.  */
+   Record the preferred register classes from the constraint for OP
+   if OP is a register.  If OP is a memory reference, record suitable
+   preferences for registers used in the address.  */
 
 void
-reg_class_record (op, constraint_loc)
+reg_class_record (op, opno, constraints)
      rtx op;
-     char **constraint_loc;
+     int opno;
+     char **constraints;
 {
-  char *constraint = *constraint_loc;
+  char *constraint = constraints[opno];
   register char *p;
   register enum reg_class class = NO_REGS;
   char *next = 0;
-  int insn_code = (constraint_loc - insn_operand_constraint[0]) / MAX_RECOG_OPERANDS;
   int memok = 0;
   int double_cost = 0;
 
@@ -499,6 +522,7 @@ reg_class_record (op, constraint_loc)
 	case '=':
 	case '?':
 	case '#':
+	case '&':
 	case '!':
 	case '%':
 	case 'F':
@@ -540,7 +564,7 @@ reg_class_record (op, constraint_loc)
 	case '4':
 	  /* If constraint says "match another operand",
 	     use that operand's constraint to choose preferences.  */
-	  next = insn_operand_constraint[insn_code][*p - '0'];
+	  next = constraints[*p - '0'];
 	  break;
 
 	default:
@@ -573,6 +597,7 @@ reg_class_record (op, constraint_loc)
 
     if (! memok)
       pp->memcost += 1 + 2 * double_cost;
+    pp->nrefs++;
   }
 }
 
@@ -673,6 +698,7 @@ record_address_regs (x, bcost, icost)
 	register struct savings *pp;
 	register enum reg_class class, class1;
 	pp = &savings[REGNO (x)];
+	pp->nrefs++;
 
 	/* We have an address (or part of one) that is just one register.  */
 
@@ -741,6 +767,12 @@ short *regno_first_uid;
 
 short *regno_last_uid;
 
+/* Maximum number of parallel sets and clobbers in any insn in this fn.
+   Always at least 3, since the combiner could put that many togetherm
+   and we want this to remain correct for all the remaining passes.  */
+
+int max_parallel;
+
 void reg_scan_mark_refs ();
 
 void
@@ -759,11 +791,18 @@ reg_scan (f, nregs, repeat)
     regno_last_uid = (short *) oballoc (nregs * sizeof (short));
   bzero (regno_last_uid, nregs * sizeof (short));
 
+  max_parallel = 3;
+
   for (insn = f; insn; insn = NEXT_INSN (insn))
     if (GET_CODE (insn) == INSN
 	|| GET_CODE (insn) == CALL_INSN
 	|| GET_CODE (insn) == JUMP_INSN)
-      reg_scan_mark_refs (PATTERN (insn), INSN_UID (insn));
+      {
+	if (GET_CODE (PATTERN (insn)) == PARALLEL
+	    && XVECLEN (PATTERN (insn), 0) > max_parallel)
+	  max_parallel = XVECLEN (PATTERN (insn), 0);
+	reg_scan_mark_refs (PATTERN (insn), INSN_UID (insn));
+      }
 }
 
 void
@@ -782,6 +821,8 @@ reg_scan_mark_refs (x, uid)
     case PC:
     case SYMBOL_REF:
     case LABEL_REF:
+    case ADDR_VEC:
+    case ADDR_DIFF_VEC:
       return;
 
     case REG:
@@ -802,7 +843,7 @@ reg_scan_mark_refs (x, uid)
 	  {
 	    if (fmt[i] == 'e')
 	      reg_scan_mark_refs (XEXP (x, i), uid);
-	    else if (fmt[i] == 'E')
+	    else if (fmt[i] == 'E' && XVEC (x, i) != 0)
 	      {
 		register int j;
 		for (j = XVECLEN (x, i) - 1; j >= 0; j--)
