@@ -75,6 +75,7 @@ actually doing the reloads, not when just counting them.
 #include "reload.h"
 #include "regs.h"
 #include "hard-reg-set.h"
+#include "flags.h"
 
 /* The variables set up by `find_reloads' are:
 
@@ -162,6 +163,10 @@ static int hard_regs_live_known;
    and is not changed here.  */
 static short *static_reload_reg_p;
 
+/* Set to 1 in subst_reg_equivs if it changes anything.  */
+static int subst_reg_equivs_changed;
+
+static int alternative_allows_memconst ();
 static rtx find_dummy_reload ();
 static rtx find_reloads_toplev ();
 static void find_reloads_address ();
@@ -203,7 +208,7 @@ push_reload (in, out, inloc, outloc, class,
      int optional;
 {
   register int i;
-  int noshare = 0;
+  int dont_share = 0;
 
   /* Compare two RTX's.  */
 #define MATCHES(x, y) (x == y || (x != 0 && GET_CODE (x) != REG && rtx_equal_p (x, y)))
@@ -244,17 +249,27 @@ push_reload (in, out, inloc, outloc, class,
 	out = gen_rtx (MEM, GET_MODE (out), XEXP (XEXP (out, 0), 0));
     }
 
+  /* If we are reloading a (SUBREG (MEM ...) ...),
+     really reload just the MEM in the MEM's own mode.  */
+
+  if (in != 0 && GET_CODE (in) == SUBREG && GET_CODE (SUBREG_REG (in)) == MEM)
+    {
+      inloc = &SUBREG_REG (in);
+      in = *inloc;
+      inmode = GET_MODE (in);
+    }
+
   /* If IN appears in OUT, we can't share any input-only reload for IN.  */
   if (in != 0 && out != 0 && reg_mentioned_p (in, out))
-    noshare = 1;
-
-  if (class == NO_REGS)
-    abort ();
+    dont_share = 1;
 
   /* Narrow down the class of register wanted if that is
      desirable on this machine for efficiency.  */
   if (in != 0)
-    class = PREFERRED_RELOAD_CLASS(in, class);
+    class = PREFERRED_RELOAD_CLASS (in, class);
+
+  if (class == NO_REGS)
+    abort ();
 
   /* We can use an existing reload if the class is right
      and at least one of IN and OUT is a match
@@ -263,7 +278,7 @@ push_reload (in, out, inloc, outloc, class,
   for (i = 0; i < n_reloads; i++)
     if (reload_reg_class[i] == class
 	&& reload_strict_low[i] == strict_low
-	&& ((in != 0 && MATCHES (reload_in[i], in) && ! noshare
+	&& ((in != 0 && MATCHES (reload_in[i], in) && ! dont_share
 	     && (out == 0 || reload_out[i] == 0 || MATCHES (reload_out[i], out)))
 	    ||
 	    (out != 0 && MATCHES (reload_out[i], out)
@@ -373,7 +388,7 @@ push_reload (in, out, inloc, outloc, class,
       if (reload_reg_rtx[i] == out
 	  && (GET_CODE (in) == REG || CONSTANT_P (in))
 	  && 0 != find_equiv_reg (in, this_insn, 0, REGNO (out),
-				  static_reload_reg_p, i))
+				  static_reload_reg_p, i, inmode))
 	reload_in[i] = out;
     }
 
@@ -677,8 +692,8 @@ strict_memory_address_p (mode, addr)
 }
 
 
-/* Like rtx_equal_p except that it considers two REGs as equal
-   if they renumber to the same value and has special hacks for
+/* Like rtx_equal_p except that it allows a REG and a SUBREG to match
+   if they are the same hard reg, and has special hacks for
    autoincrement and autodecrement.
    This is specifically intended for find_reloads to use
    in determining whether two operands match.
@@ -694,7 +709,7 @@ strict_memory_address_p (mode, addr)
 
 int
 operands_match_p (x, y)
-     rtx x, y;
+     register rtx x, y;
 {
   register int i;
   register RTX_CODE code = GET_CODE (x);
@@ -708,32 +723,27 @@ operands_match_p (x, y)
 				  && GET_CODE (SUBREG_REG (y)) == REG)))
     {
       register int j;
+
       if (code == SUBREG)
 	{
 	  i = REGNO (SUBREG_REG (x));
-	  if (reg_renumber[i] >= 0)
-	    i = reg_renumber[i];
+	  if (i >= FIRST_PSEUDO_REGISTER)
+	    goto slow;
 	  i += SUBREG_WORD (x);
 	}
       else
-	{
-	  i = REGNO (x);
-	  if (reg_renumber[i] >= 0)
-	    i = reg_renumber[i];
-	}
+	i = REGNO (x);
+
       if (GET_CODE (y) == SUBREG)
 	{
 	  j = REGNO (SUBREG_REG (y));
-	  if (reg_renumber[j] >= 0)
-	    j = reg_renumber[j];
+	  if (j >= FIRST_PSEUDO_REGISTER)
+	    goto slow;
 	  j += SUBREG_WORD (y);
 	}
       else
-	{
-	  j = REGNO (y);
-	  if (reg_renumber[j] >= 0)
-	    j = reg_renumber[j];
-	}
+	j = REGNO (y);
+
       return i == j;
     }
   /* If two operands must match, because they are really a single
@@ -760,6 +770,7 @@ operands_match_p (x, y)
   if (code == SYMBOL_REF)
     return XSTR (x, 0) == XSTR (y, 0);
 
+ slow:
   /* (MULT:SI x y) and (MULT:HI x y) are NOT equivalent.  */
 
   if (GET_MODE (x) != GET_MODE (y))
@@ -815,7 +826,7 @@ operands_match_p (x, y)
    allocation has been done.
 
    RELOAD_REG_P if nonzero is a vector indexed by hard reg number
-   which is nonzero if the reg has been commandeered for reloading into.
+   which is nonnegative if the reg has been commandeered for reloading into.
    It is copied into STATIC_RELOAD_REG_P and referenced from there
    by various subroutines.  */
 
@@ -888,8 +899,15 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
     case ADDR_DIFF_VEC:
       return;
 
-    case PARALLEL:
     case SET:
+      /* Dispose quickly of (set (reg..) (reg..)) if both have hard regs.  */
+      if (GET_CODE (SET_DEST (body)) == REG
+	  && REGNO (SET_DEST (body)) < FIRST_PSEUDO_REGISTER
+	  && GET_CODE (SET_SRC (body)) == REG
+	  && REGNO (SET_SRC (body)) < FIRST_PSEUDO_REGISTER)
+	return;
+    case PARALLEL:
+    case ASM_OPERANDS:
       noperands = asm_noperands (body);
       if (noperands > 0)
 	{
@@ -916,12 +934,23 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
       insn_code_number = recog_memoized (insn);
       noperands = insn_n_operands[insn_code_number];
       insn_extract (insn);
-      for (i = 0; i < noperands; i++)
-	{
-	  constraints[i] = constraints1[i]
-	    = insn_operand_constraint[insn_code_number][i];
-	  operand_mode[i] = insn_operand_mode[insn_code_number][i];
-	}
+      {
+	/* Nonzero if some operand has a nonnull constraint.
+	   If none has, we have no work to do.
+	   Even a general_operand should have `g' in its constraint.  */
+	int have_constraints = 0;
+
+	for (i = 0; i < noperands; i++)
+	  {
+	    constraints[i] = constraints1[i]
+	      = insn_operand_constraint[insn_code_number][i];
+	    operand_mode[i] = insn_operand_mode[insn_code_number][i];
+	    if (constraints[i] != 0)
+	      have_constraints = 1;
+	  }
+	if (!have_constraints)
+	  return;
+      }
     }
 
   if (noperands == 0)
@@ -1066,25 +1095,20 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	  int c;
 	  register rtx operand = recog_operand[i];
 	  int offset = 0;
+	  /* Nonzero means this is a MEM that must be reloaded into a reg
+	     regardless of what the constraint says.  */
 	  int force_reload = 0;
 	  int offmemok = 0;
 	  int earlyclobber = 0;
 
 	  /* If the operand is a SUBREG, extract
-	     the REG or MEM within.  */
+	     the REG or MEM (or maybe CONST_DOUBLE) within.  */
 
 	  while (GET_CODE (operand) == SUBREG)
 	    {
 	      offset += SUBREG_WORD (operand);
 	      operand = SUBREG_REG (operand);
-	      if (GET_CODE (operand) == MEM
-/*** This is overcautious, as for BYTES_BIG_ENDIAN it is still possible
-     to avoid setting force_reload if the mode of the subreg
-     is SImode or bigger.  */
-#ifndef BYTES_BIG_ENDIAN
-		  && offset != 0
-#endif
-		  && !offsetable_memref_p (operand))
+	      if (GET_CODE (operand) == MEM)
 		force_reload = 1;
 	    }
 
@@ -1193,12 +1217,15 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 		break;
 
 	      case 'm':
+		if (force_reload)
+		  break;
 		if (GET_CODE (operand) == MEM
 		    || (GET_CODE (operand) == REG
 			&& REGNO (operand) >= FIRST_PSEUDO_REGISTER
 			&& reg_renumber[REGNO (operand)] < 0))
 		  win = 1;
-		if (GET_CODE (operand) == CONST_DOUBLE)
+		if (GET_CODE (operand) == CONST_DOUBLE
+		    || CONSTANT_P (operand))
 		  badop = 0;
 		break;
 
@@ -1218,6 +1245,8 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 
 		/* Memory operand whose address is offsettable.  */
 	      case 'o':
+		if (force_reload)
+		  break;
 		if ((GET_CODE (operand) == MEM
 		     && offsetable_memref_p (operand))
 		    || (GET_CODE (operand) == REG
@@ -1225,6 +1254,7 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 			&& reg_renumber[REGNO (operand)] < 0))
 		  win = 1;
 		if (GET_CODE (operand) == CONST_DOUBLE
+		    || CONSTANT_P (operand)
 		    || (GET_CODE (operand) == MEM
 			&& GET_CODE (XEXP (operand, 0)) != POST_INC
 			&& GET_CODE (XEXP (operand, 0)) != POST_DEC
@@ -1277,10 +1307,11 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 		break;
 
 	      case 'g':
-		if (GENERAL_REGS == ALL_REGS
-		    || GET_CODE (operand) != REG
-		    || (REGNO (operand) >= FIRST_PSEUDO_REGISTER
-			&& reg_renumber[REGNO (operand)] < 0))
+		if (! force_reload
+		    && (GENERAL_REGS == ALL_REGS
+			|| GET_CODE (operand) != REG
+			|| (REGNO (operand) >= FIRST_PSEUDO_REGISTER
+			    && reg_renumber[REGNO (operand)] < 0)))
 		  win = 1;
 		/* Drop through into 'r' case */
 
@@ -1292,7 +1323,7 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	      default:
 		this_alternative[i]
 		  = (int) reg_class_subunion[this_alternative[i]][(int) REG_CLASS_FROM_LETTER (c)];
-
+		
 	      reg:
 		badop = 0;
 		if (GET_CODE (operand) == REG
@@ -1314,6 +1345,11 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	      losers++;
 	      if (badop)
 		bad = 1;
+	      /* Alternative loses if it has no regs for a reg operand.  */
+	      if (GET_CODE (operand) == REG
+		  && this_alternative[i] == (int) NO_REGS
+		  && this_alternative_matches[i] < 0)
+		bad = 1;
 	    }
 	}
 
@@ -1326,6 +1362,7 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	    && this_alternative_win[i])
 	  {
 	    int j;
+
 	    for (j = 0; j < noperands; j++)
 	      /* Is this an input operand or a memory ref?  */
 	      if ((GET_CODE (recog_operand[j]) == MEM
@@ -1333,7 +1370,18 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 		  /* Does it refer to the earlyclobber operand?  */
 		  && refers_to_regno_p (REGNO (recog_operand[i]),
 					recog_operand[j], 0))
-		break;
+		{
+		  /* If the output is in a single-reg class,
+		     it's costly to reload it, so reload the input instead.  */
+		  if (reg_class_size[this_alternative[i]] == 1
+		      && GET_CODE (recog_operand[j]) == REG)
+		    {
+		      losers++;
+		      this_alternative_win[j] = 0;
+		    }
+		  else
+		    break;
+		}
 	    /* If an earlyclobber operand conflicts with something,
 	       it must be reloaded, so request this and count the cost.  */
 	    if (j != noperands)
@@ -1351,13 +1399,12 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	  recog_operand[1] = substed_operand[1];
 	  recog_operand[2] = substed_operand[2];
 	  for (i = 0; i < noperands; i++)
-	    goal_alternative_win[i] = 1;
-	  bcopy (this_alternative, goal_alternative,
-		 sizeof this_alternative);
-	  bcopy (this_alternative_offmemok, goal_alternative_offmemok,
-		 sizeof this_alternative_offmemok);
-	  bcopy (this_alternative_matches, goal_alternative_matches,
-		 sizeof this_alternative_matches);
+	    {
+	      goal_alternative_win[i] = 1;
+	      goal_alternative[i] = this_alternative[i];
+	      goal_alternative_offmemok[i] = this_alternative_offmemok[i];
+	      goal_alternative_matches[i] = this_alternative_matches[i];
+	    }
 	  goal_alternative_number = this_alternative_number;
 	  goal_alternative_swapped = swapped;
 	  goal_earlyclobber = this_earlyclobber;
@@ -1374,14 +1421,13 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	 record it as the chosen goal for reloading.  */
       if (! bad && best > losers)
 	{
-	  bcopy (this_alternative, goal_alternative,
-		 sizeof this_alternative);
-	  bcopy (this_alternative_win, goal_alternative_win,
-		 sizeof this_alternative_win);
-	  bcopy (this_alternative_offmemok, goal_alternative_offmemok,
-		 sizeof this_alternative_offmemok);
-	  bcopy (this_alternative_matches, goal_alternative_matches,
-		 sizeof this_alternative_matches);
+	  for (i = 0; i < noperands; i++)
+	    {
+	      goal_alternative[i] = this_alternative[i];
+	      goal_alternative_win[i] = this_alternative_win[i];
+	      goal_alternative_offmemok[i] = this_alternative_offmemok[i];
+	      goal_alternative_matches[i] = this_alternative_matches[i];
+	    }
 	  goal_alternative_swapped = swapped;
 	  best = losers;
 	  goal_alternative_number = this_alternative_number;
@@ -1423,7 +1469,14 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
      Reload so as to fit it.  */
 
   if (best == MAX_RECOG_OPERANDS + 100)
-    abort ();			/* No alternative works with reloads??  */
+    {
+      /* No alternative works with reloads??  */
+      if (insn_code_number >= 0)
+	abort ();
+      error ("inconsistent operand constraints in an `asm' in this function");
+      n_reloads = 0;
+      return;
+    }
 
   /* Jump to `finish' from above if all operands are valid already.
      In that case, goal_alternative_win is all 1.  */
@@ -1469,6 +1522,27 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
       operand_reloadnum[i] = -1;
     }
 
+  /* Any floating constants that aren't allowed and can't be reloaded
+     into memory locations are here changed into memory references.  */
+  for (i = 0; i < noperands; i++)
+    if (! goal_alternative_win[i]
+	&& (GET_CODE (recog_operand[i]) == CONST_DOUBLE
+	    || CONSTANT_P (recog_operand[i]))
+	&& (PREFERRED_RELOAD_CLASS (recog_operand[i],
+				    (enum reg_class) goal_alternative[i])
+	    == NO_REGS))
+      {
+	enum machine_mode mode = GET_MODE (recog_operand[i]);
+	*recog_operand_loc[i] = recog_operand[i]
+	  = (GET_CODE (recog_operand[i]) == CONST_DOUBLE
+	     ? force_const_double_mem (recog_operand[i])
+	     : force_const_mem (mode != VOIDmode ? mode : SImode,
+				recog_operand[i]));
+	find_reloads_toplev (recog_operand[i]);
+	if (alternative_allows_memconst (constraints1[i], goal_alternative_number))
+	  goal_alternative_win[i] = 1;
+      }
+
   /* Now record reloads for all the operands that need them.  */
   for (i = 0; i < noperands; i++)
     if (! goal_alternative_win[i])
@@ -1476,19 +1550,6 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	/* Operands that match previous ones have already been handled.  */
 	if (goal_alternative_matches[i] >= 0)
 	  ;
-	/* This clause forces a double constant into memory
-	   if necessary.  But right now it appears never necessary.
-	   Perhaps there should be a heuristic here to detect cases
-	   when it is desirable, even though not necessary, to move
-	   the constant to memory.  I can't decide when it is desirable.  */
-	else if (GET_CODE (recog_operand[i]) == CONST_DOUBLE
-		 && alternative_allows_memconst (constraints1[i], goal_alternative_number)
-		 && goal_alternative[i] == (int) NO_REGS)
-	  {
-	    *recog_operand_loc[i] = recog_operand[i]
-	      = force_const_double_mem (recog_operand[i]);
-	    find_reloads_toplev (recog_operand[i]);
-	  }
 	/* Handle an operand with a nonoffsetable address
 	   appearing where an offsetable address will do
 	   by reloading the address into a base register.  */
@@ -1557,7 +1618,11 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
 	if (GET_CODE (operand) == REG
 	    && REGNO (operand) >= FIRST_PSEUDO_REGISTER
 	    && reg_renumber[REGNO (operand)] < 0
-	    && (enum reg_class) goal_alternative[i] != NO_REGS)
+	    && (enum reg_class) goal_alternative[i] != NO_REGS
+	    /* Don't make optional output reloads for jump insns
+	       (such as aobjeq on the vax).  */
+	    && (modified[i] == RELOAD_READ
+		|| GET_CODE (insn) != JUMP_INSN))
 	  operand_reloadnum[i]
 	    = push_reload (modified[i] != RELOAD_WRITE ? recog_operand[i] : 0,
 			   modified[i] != RELOAD_READ ? recog_operand[i] : 0,
@@ -1600,7 +1665,7 @@ find_reloads (insn, replace, ind_ok, live_known, reload_reg_p)
       {
 	reload_reg_rtx[i]
 	  = find_equiv_reg (reload_in[i], insn, reload_reg_class[i], -1,
-			    static_reload_reg_p, 0);
+			    static_reload_reg_p, 0, reload_inmode[i]);
 	/* Prevent generation of insn to load the value
 	   because the one we found already has the value.  */
 	if (reload_reg_rtx[i])
@@ -1837,8 +1902,19 @@ find_reloads_address (mode, memrefloc, ad, loc)
 	 it has failed to get a hard reg.
 	 So do a tree-walk to find and eliminate all such regs.  */
 
+      /* But first quickly dispose of a common case.  */
+      if (GET_CODE (ad) == PLUS
+	  && GET_CODE (XEXP (ad, 1)) == CONST_INT
+	  && GET_CODE (XEXP (ad, 0)) == REG
+	  && reg_equiv_constant[REGNO (XEXP (ad, 0))] == 0)
+	return;
+
+      subst_reg_equivs_changed = 0;
       *loc = subst_reg_equivs (ad);
 
+      if (! subst_reg_equivs_changed)
+	return;
+
       /* Check result for validity after substitution.  */
       if (strict_memory_address_p (mode, ad))
 	return;
@@ -1847,9 +1923,11 @@ find_reloads_address (mode, memrefloc, ad, loc)
   /* If we have address of a stack slot but it's not valid
      (displacement is too large), compute the sum in a register.  */
   if (GET_CODE (ad) == PLUS
-      && GET_CODE (XEXP (ad, 0)) == REG
-      && (REGNO (XEXP (ad, 0)) == FRAME_POINTER_REGNUM
-	  || REGNO (XEXP (ad, 0)) == ARG_POINTER_REGNUM)
+      && (XEXP (ad, 0) == frame_pointer_rtx
+#if FRAME_POINTER_REGNUM != ARG_POINTER_REGNUM
+	  || XEXP (ad, 0) == arg_pointer_rtx
+#endif
+	  )
       && GET_CODE (XEXP (ad, 1)) == CONST_INT)
     {
       /* Unshare the MEM rtx so we can safely alter it.  */
@@ -1866,17 +1944,20 @@ find_reloads_address (mode, memrefloc, ad, loc)
   /* See if address becomes valid when an eliminable register
      in a sum is replaced.  */
 
-  tem = subst_indexed_address (ad);
+  tem = ad;
+  if (GET_CODE (ad) == PLUS)
+    tem = subst_indexed_address (ad);
   if (tem != ad && strict_memory_address_p (mode, tem))
     {
       /* Ok, we win that way.  Replace any additional eliminable
 	 registers.  */
 
+      subst_reg_equivs_changed = 0;
       tem = subst_reg_equivs (tem);
 
       /* Make sure that didn't make the address invalid again.  */
 
-      if (strict_memory_address_p (mode, tem))
+      if (! subst_reg_equivs_changed || strict_memory_address_p (mode, tem))
 	{
 	  *loc = tem;
 	  return;
@@ -1923,11 +2004,19 @@ subst_reg_equivs (ad)
       {
 	register int regno = REGNO (ad);
 
-	if (regno >= FIRST_PSEUDO_REGISTER && reg_renumber[regno] < 0
-	    && reg_equiv_constant[regno] != 0)
-	  return reg_equiv_constant[regno];
+	if (reg_equiv_constant[regno] != 0)
+	  {
+	    subst_reg_equivs_changed = 1;
+	    return reg_equiv_constant[regno];
+	  }
       }
       return ad;
+
+    case PLUS:
+      /* Quickly dispose of a common case.  */
+      if (XEXP (ad, 0) == frame_pointer_rtx
+	  && GET_CODE (XEXP (ad, 1)) == CONST_INT)
+	return ad;
     }
 
   fmt = GET_RTX_FORMAT (code);
@@ -1936,7 +2025,7 @@ subst_reg_equivs (ad)
       XEXP (ad, i) = subst_reg_equivs (XEXP (ad, i));
   return ad;
 }
-
+
 /* If ADDR is a sum containing a pseudo register that should be
    replaced with a constant (from reg_equiv_constant),
    return the result of doing so, and also apply the associative
@@ -1961,20 +2050,21 @@ subst_indexed_address (addr)
       else if (CONSTANT_P (XEXP (addr, 1)))
 	const_part = XEXP (addr, 1),
 	var_part = XEXP (addr, 0);
+      else
+	var_part = addr;
 
-      if (const_part == 0)
-	return addr;
-
-      if (GET_CODE (const_part) == CONST)
+      if (const_part && GET_CODE (const_part) == CONST)
 	const_part = XEXP (const_part, 0);
 
       if (GET_CODE (var_part) == REG
 	  && (regno = REGNO (var_part)) >= FIRST_PSEUDO_REGISTER
 	  && reg_renumber[regno] < 0
 	  && reg_equiv_constant[regno] != 0)
-	return gen_rtx (CONST, VOIDmode,
-			gen_rtx (PLUS, Pmode, const_part,
-				 reg_equiv_constant[regno]));
+	return (const_part
+	        ? gen_rtx (CONST, VOIDmode,
+			   gen_rtx (PLUS, Pmode, const_part,
+				    reg_equiv_constant[regno]))
+		: reg_equiv_constant[regno]);
 
       if (GET_CODE (var_part) != PLUS)
 	return addr;
@@ -1984,18 +2074,22 @@ subst_indexed_address (addr)
 	  && reg_renumber[regno] < 0
 	  && reg_equiv_constant[regno] != 0)
 	return gen_rtx (PLUS, Pmode, XEXP (var_part, 1),
-			gen_rtx (CONST, VOIDmode,
-				 gen_rtx (PLUS, Pmode, const_part,
-					  reg_equiv_constant[regno])));
+			(const_part
+			 ? gen_rtx (CONST, VOIDmode,
+				    gen_rtx (PLUS, Pmode, const_part,
+					     reg_equiv_constant[regno]))
+			 : reg_equiv_constant[regno]));
 
       if (GET_CODE (XEXP (var_part, 1)) == REG
 	  && (regno = REGNO (XEXP (var_part, 1))) >= FIRST_PSEUDO_REGISTER
 	  && reg_renumber[regno] < 0
 	  && reg_equiv_constant[regno] != 0)
 	return gen_rtx (PLUS, Pmode, XEXP (var_part, 0),
-			gen_rtx (CONST, VOIDmode,
-				 gen_rtx (PLUS, Pmode, const_part,
-					  reg_equiv_constant[regno])));
+			(const_part
+			 ? gen_rtx (CONST, VOIDmode,
+				    gen_rtx (PLUS, Pmode, const_part,
+					     reg_equiv_constant[regno]))
+			 : reg_equiv_constant[regno]));
     }
   return addr;
 }
@@ -2288,18 +2382,22 @@ forget_volatility (x)
    If GOAL is zero, then GOALREG is a register number; we look
    for an equivalent for that register.
 
+   MODE is the machine mode of the value we want an equivalence for.
+   If GOAL is nonzero and not VOIDmode, then it must have mode MODE.
+
    This function is used by jump.c as well as in the reload pass.
 
    If GOAL is a PLUS, we assume it adds the stack pointer to a constant.  */
 
 rtx
-find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
+find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg, mode)
      register rtx goal;
      rtx insn;
      enum reg_class class;
      register int other;
      short *reload_reg_p;
      int goalreg;
+     enum machine_mode mode;
 {
   register rtx p = insn;
   rtx valtry, value, where;
@@ -2308,18 +2406,46 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
   int valueno;
   int goal_mem = 0;
   int goal_const = 0;
+  int goal_mem_addr_varies = 0;
+  int nregs;
+  int valuenregs;
 
   if (goal == 0)
     regno = goalreg;
   else if (GET_CODE (goal) == REG)
     regno = REGNO (goal);
   else if (GET_CODE (goal) == MEM)
-    goal_mem = 1;
+    {
+      enum rtx_code code = GET_CODE (XEXP (goal, 0));
+      if (goal->volatil)
+	return 0;
+      if (flag_float_store
+	  && (GET_MODE (goal) == DFmode || GET_MODE (goal) == SFmode))
+	return 0;
+      /* An address with side effects must be reexecuted.  */
+      switch (code)
+	{
+	case POST_INC:
+	case PRE_INC:
+	case POST_DEC:
+	case PRE_DEC:
+	  return 0;
+	}
+      goal_mem = 1;
+    }
   else if (CONSTANT_P (goal))
     goal_const = 1;
   else
     return 0;
 
+  /* On some machines, certain regs must always be rejected
+     because they don't behave the way ordinary registers do.  */
+  
+#ifdef OVERLAPPING_REGNO_P
+   if (regno >= 0 && OVERLAPPING_REGNO_P (regno))
+     return 0;
+#endif      
+
   /* Scan insns back from INSN, looking for one that copies
      a value into or out of GOAL.
      Stop and give up if we reach a label.  */
@@ -2346,12 +2472,12 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 	  if (GET_CODE (pat) == SET
 	      && ((regno >= 0
 		   && GET_CODE (SET_SRC (pat)) == REG
-		   && (goal == 0 ? true_regnum (SET_SRC (pat)) : REGNO (SET_SRC (pat))) == regno
+		   && REGNO (SET_SRC (pat)) == regno
 		   && GET_CODE (valtry = SET_DEST (pat)) == REG)
 		  ||
 		  (regno >= 0
 		   && GET_CODE (SET_DEST (pat)) == REG
-		   && (goal == 0 ? true_regnum (SET_DEST (pat)) : REGNO (SET_DEST (pat))) == regno
+		   && REGNO (SET_DEST (pat)) == regno
 		   && GET_CODE (valtry = SET_SRC (pat)) == REG)
 		  ||
 		  (goal_const && rtx_equal_p (SET_SRC (pat), goal)
@@ -2362,13 +2488,12 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 		  || (goal_mem
 		      && GET_CODE (valtry = SET_SRC (pat)) == REG
 		      && rtx_renumbered_equal_p (goal, SET_DEST (pat)))))
-	    if (other >= 0
-		? (goal == 0 ? true_regnum (valtry) : REGNO (valtry)) == other
-		: (valueno = REGNO (valtry),
-		   reg_renumber[valueno] >= 0 ? valueno = reg_renumber[valueno] : 0,
-		   valueno < FIRST_PSEUDO_REGISTER &&
-		   TEST_HARD_REG_BIT (reg_class_contents[(int) class],
-				      valueno)))
+	    if (valueno = REGNO (valtry),
+		other >= 0
+		? valueno == other
+		: ((unsigned) valueno < FIRST_PSEUDO_REGISTER
+		   && TEST_HARD_REG_BIT (reg_class_contents[(int) class],
+					 valueno)))
 	      {
 		value = valtry;
 		where = p;
@@ -2381,25 +2506,17 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
      (or copying VALUE into GOAL, if GOAL is also a register).
      Now verify that VALUE is really valid.  */
 
-  if (regno >= FIRST_PSEUDO_REGISTER && reg_renumber[regno] >= 0)
-    regno = reg_renumber[regno];
-
-  /* VALUENO gets the register number of VALUE;
-     for a pseudo reg, it gets the hard reg number that the pseudo has,
-     and we give up if the pseudo has no hard reg.  */
-
-  valueno = REGNO (value);
-  if (valueno >= FIRST_PSEUDO_REGISTER)
-    valueno = reg_renumber[valueno];
-
-  if (valueno < 0)
-    return 0;
+  /* VALUENO is the register number of VALUE; a hard register.  */
 
   /* Don't find the sp as an equiv, since pushes that we don't notice
      would invalidate it.  */
   if (valueno == STACK_POINTER_REGNUM)
     return 0;
 
+  /* Reject VALUE if the copy-insn moved the wrong sort of datum.  */
+  if (GET_MODE (value) != mode)
+    return 0;
+
   /* Reject VALUE if it was loaded from GOAL
      and is also a register that appears in the address of GOAL.  */
 
@@ -2417,6 +2534,17 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
       && reload_reg_p[valueno] >= 0)
     return 0;
 
+  /* On some machines, certain regs must always be rejected
+     because they don't behave the way ordinary registers do.  */
+  
+#ifdef OVERLAPPING_REGNO_P
+  if (OVERLAPPING_REGNO_P (valueno))
+    return 0;
+#endif      
+
+  nregs = HARD_REGNO_NREGS (regno, mode);
+  valuenregs = HARD_REGNO_NREGS (valueno, mode);
+
   /* Reject VALUE if it is a register being used for an input reload
      even if it is not one of those reserved.  */
 
@@ -2427,13 +2555,17 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 	if (reload_reg_rtx[i] != 0 && reload_in[i])
 	  {
 	    int regno1 = REGNO (reload_reg_rtx[i]);
-	    if (reg_renumber[regno1] >= 0)
-	      regno1 = reg_renumber[regno1];
-	    if (valueno == regno1)
+	    int nregs1 = HARD_REGNO_NREGS (regno1,
+					   GET_MODE (reload_reg_rtx[i]));
+	    if (regno1 < valueno + valuenregs
+		&& regno1 + nregs1 > valueno)
 	      return 0;
 	  }
     }
 
+  if (goal_mem)
+    goal_mem_addr_varies = rtx_addr_varies_p (goal);
+
   /* Now verify that the values of GOAL and VALUE remain unaltered
      until INSN is reached.  */
 
@@ -2477,9 +2609,13 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 	      if (GET_CODE (dest) == REG)
 		{
 		  register int xregno = REGNO (dest);
-		  if (reg_renumber[xregno] >= 0)
-		    xregno = reg_renumber[xregno];
-		  if (xregno == regno || xregno == valueno || goal_mem)
+		  int xnregs = HARD_REGNO_NREGS (xregno, GET_MODE (dest));
+		  if (xregno < regno + nregs && xregno + xnregs > regno)
+		    return 0;
+		  if (xregno < valueno + valuenregs
+		      && xregno + xnregs > valueno)
+		    return 0;
+		  if (goal_mem_addr_varies)
 		    return 0;
 		}
 	      else if (goal_mem && GET_CODE (dest) == MEM
@@ -2503,9 +2639,14 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 		      if (GET_CODE (dest) == REG)
 			{
 			  register int xregno = REGNO (dest);
-			  if (reg_renumber[xregno] >= 0)
-			    xregno = reg_renumber[xregno];
-			  if (xregno == regno || xregno == valueno || goal_mem)
+			  int xnregs = HARD_REGNO_NREGS (xregno, GET_MODE (dest));
+			  if (xregno < regno + nregs
+			      && xregno + xnregs > regno)
+			    return 0;
+			  if (xregno < valueno + valuenregs
+			      && xregno + xnregs > valueno)
+			    return 0;
+			  if (goal_mem_addr_varies)
 			    return 0;
 			}
 		      else if (goal_mem && GET_CODE (dest) == MEM
@@ -2526,9 +2667,11 @@ find_equiv_reg (goal, insn, class, other, reload_reg_p, goalreg)
 	      if (REG_NOTE_KIND (link) == REG_INC)
 		{
 		  register int incno = REGNO (XEXP (link, 0));
-		  if (reg_renumber[incno] >= 0)
-		    incno = reg_renumber[incno];
-		  if (incno == regno || incno == valueno || goal_mem)
+		  if (incno < regno + nregs && incno >= regno)
+		    return 0;
+		  if (incno < valueno + valuenregs && incno >= valueno)
+		    return 0;
+		  if (goal_mem_addr_varies)
 		    return 0;
 		}
 	  }
