@@ -126,6 +126,46 @@ get_temporary_types ()
   return tem;
 }
 
+/* SAVE_EXPRs for sizes of types and decls, waiting to be expanded.  */
+
+static tree pending_sizes;
+
+/* Nonzero means cannot safely call expand_expr now,
+   so put variable sizes onto `pending_sizes' instead.  */
+
+int immediate_size_expand;
+
+tree
+get_pending_sizes ()
+{
+  tree chain = pending_sizes;
+  pending_sizes = 0;
+  return chain;
+}
+
+/* Given a size SIZE that isn't constant, return a SAVE_EXPR
+   to serve as the actual size-expression for a type or decl.  */
+
+static tree
+variable_size (size)
+     tree size;
+{
+  size = save_expr (size);
+
+  if (global_bindings_p ())
+    {
+      error ("variable-size type declared outside of any function");
+      return build_int (1);
+    }
+
+  if (immediate_size_expand)
+    expand_expr (size, 0, VOIDmode, 0);
+  else
+    pending_sizes = tree_cons (0, size, pending_sizes);
+
+  return size;
+}
+
 /* Return the machine mode to use for an aggregate of SIZE bits.
 
    Note!!!  We only use a non-BLKmode mode if the size matches exactly.
@@ -220,30 +260,6 @@ convert_units (size, inunits, outunits)
   return genop (CEIL_DIV_EXPR, t,
 		build_int (outunits)); /* then to outunits */
 }
-
-/* Form a tree giving the value of
-   CEIL ((CONST + VAR * VAR_UNIT) / BITS_PER_UNIT)
-   where VAR is any expression and CONST and VAR_UNIT are integers.
-   VAR may be null; that counts as zero.  */
-
-static
-tree
-add_vc_sizes (constant, var, coeff)
-     int constant;
-     tree var;
-     int coeff;
-{
-  register tree tmp1;
-
-  if (var == 0)
-    return build_int (CEIL (constant, BITS_PER_UNIT));
-  if (constant == 0)
-    return convert_units (var, coeff, BITS_PER_UNIT);
-
-  tmp1 = genop (PLUS_EXPR, genop (MULT_EXPR, var, size_one_node),
-		build_int (constant)); /* add */
-  return genop (CEIL_DIV_EXPR, tmp1, build_int (BITS_PER_UNIT));
-}
 
 /* Set the size, mode and alignment of a ..._DECL node.
    Note that LABEL_DECL, TYPE_DECL and CONST_DECL nodes do not need this,
@@ -266,8 +282,11 @@ layout_decl (decl, known_align)
   register enum tree_code code = TREE_CODE (decl);
   int spec_size = DECL_SIZE_UNIT (decl);
 
+  if (code == CONST_DECL)
+    return;
+
   if (code != VAR_DECL && code != PARM_DECL && code != RESULT_DECL
-      && code != FIELD_DECL)
+      && code != FIELD_DECL && code != FRIEND_DECL)
     abort ();
 
   if (type == error_mark_node)
@@ -315,6 +334,8 @@ layout_decl (decl, known_align)
      and occupying a complete byte or bytes on proper boundary.  */
   if ((DECL_MODE (decl) == BLKmode
        || DECL_MODE (decl) == BImode)
+      /* Don't do this if DECL's type requires it to be BLKmode.  */
+      && TYPE_MODE (type) != BLKmode
       && TYPE_SIZE (type) != 0
       && TREE_CODE (TYPE_SIZE (type)) == INTEGER_CST)
     {
@@ -332,13 +353,17 @@ layout_decl (decl, known_align)
 	  DECL_SIZE_UNIT (decl) = BITS_PER_UNIT;
 	}
     }
+
+  /* Evaluate nonconstant size only once, either now or as soon as safe.  */
+  if (DECL_SIZE (decl) != 0 && ! TREE_LITERAL (DECL_SIZE (decl)))
+    DECL_SIZE (decl) = variable_size (DECL_SIZE (decl));
 }
 
 /* Lay out a RECORD_TYPE type (a C struct).
    This means laying out the fields, determining their offsets,
    and computing the overall size and required alignment of the record.
    Note that if you set the TYPE_ALIGN before calling this
-   then the union align is aligned to at least that boundary.  */
+   then the struct is aligned to at least that boundary.  */
 
 static void
 layout_record (rec)
@@ -346,10 +371,12 @@ layout_record (rec)
 {
   register tree field;
 #ifdef STRUCTURE_SIZE_BOUNDARY
-  int record_align = MIN (STRUCTURE_SIZE_BOUNDARY, TYPE_ALIGN (rec));
+  int record_align = MAX (STRUCTURE_SIZE_BOUNDARY, TYPE_ALIGN (rec));
 #else
-  int record_align = MIN (BITS_PER_UNIT, TYPE_ALIGN (rec));
+  int record_align = MAX (BITS_PER_UNIT, TYPE_ALIGN (rec));
 #endif
+  /* These must be laid out *after* the record is.  */
+  tree pending_statics = NULL_TREE;
   /* Record size so far is CONST_SIZE + VAR_SIZE * SIZE_UNIT bits,
      where CONST_SIZE is an integer
      and VAR_SIZE is a tree expression.
@@ -359,10 +386,55 @@ layout_record (rec)
   register tree var_size = 0;
   register int size_unit = BITS_PER_UNIT;
 
+  int n_basetypes = list_length (TYPE_BASETYPES (rec));
+
+  /* Handle basetypes almost like fields, but record their
+     offsets differently.  */
+
+  for (field = TYPE_BASETYPES (rec); field; field = TREE_CHAIN (field))
+    {
+      tree basetype = TREE_VALUE (field);
+      int desired_align = TYPE_ALIGN (basetype);
+
+      /* Record must have at least as much alignment as any basetype.  */
+      record_align = MAX (record_align, desired_align);
+
+      /* Does this basetype have the alignment it needs
+	 by virtue of the basetypes that precede it?  */
+
+      if (const_size % desired_align != 0)
+	{
+	  /* No, we need to skip space before this field.
+	     Bump the cumulative size to multiple of field alignment.  */
+
+	  const_size = CEIL (const_size, desired_align) * desired_align;
+	}
+
+      /* Record this basetype's offset within the record.  */
+      TREE_PURPOSE (field) =
+	convert_units (build_int (const_size), 1, BITS_PER_UNIT);
+
+      /* Skip enough space for this basetype.  */
+      const_size += (TREE_INT_CST_LOW (TYPE_SIZE (basetype)) *
+		     TYPE_SIZE_UNIT (basetype));
+    }
+
   for (field = TYPE_FIELDS (rec); field; field = TREE_CHAIN (field))
     {
       register int desired_align;
 
+      /* If FIELD is a VAR_DECL, then treat it like a separate variable,
+	 not really like a structure field.
+	 If it is a FUNCTION_DECL, it's a method.
+	 In both cases, all we do is lay out the decl,
+	 and we do it *after* the record is laid out.  */
+
+      if (TREE_CODE (field) != FIELD_DECL)
+	{
+	  pending_statics = tree_cons (NULL, field, pending_statics);
+	  continue;
+	}
+
       /* Lay out the field so we know what alignment it needs.
 	 For KNOWN_ALIGN, pass the number of bits from start of record
 	 or some divisor of it.  */
@@ -375,6 +447,11 @@ layout_record (rec)
 	 is meaningless.  */
 
       record_align = MAX (record_align, desired_align);
+#ifdef PCC_BITFIELD_TYPE_MATTERS
+      /* In PCC on Vax, Sony, etc., a bit field of declare type `int'
+	 forces the entire structure to have `int' alignment.  */
+      record_align = MAX (record_align, TYPE_ALIGN (TREE_TYPE (field)));
+#endif
 
       /* Does this field automatically have alignment it needs
 	 by virtue of the fields that precede it and the record's
@@ -456,6 +533,12 @@ layout_record (rec)
 				   BITS_PER_UNIT);
   TYPE_SIZE_UNIT (rec) = BITS_PER_UNIT;
   TYPE_ALIGN (rec) = MIN (BIGGEST_ALIGNMENT, record_align);
+
+  /* Lay out any static members.  This is done now
+     because their type may use the record's type.  */
+
+  for (field = pending_statics; field; field = TREE_CHAIN (field))
+    layout_decl (TREE_VALUE (field));
 }
 
 /* Lay out a UNION_TYPE type.
@@ -546,6 +629,7 @@ layout_type (type)
      tree type;
 {
   int old;
+  int temporary = 0;
 
   if (type == 0)
     abort ();
@@ -557,6 +641,11 @@ layout_type (type)
   /* Make sure all nodes we allocate are not momentary;
      they must last past the current statement.  */
   old  = suspend_momentary ();
+  if (TREE_PERMANENT (type) && allocation_temporary_p ())
+    {
+      temporary = 1;
+      end_temporary_allocation ();
+    }
 
   chain_type (type);
 
@@ -596,6 +685,7 @@ layout_type (type)
       break;
 
     case POINTER_TYPE:
+    case REFERENCE_TYPE:
       TYPE_MODE (type) = Pmode;
       TYPE_SIZE (type) = build_int (POINTER_SIZE / BITS_PER_UNIT);
       TYPE_SIZE_UNIT (type) = BITS_PER_UNIT;
@@ -645,9 +735,17 @@ layout_type (type)
       TYPE_MODE (type) = BLKmode;
       if (TREE_CODE (TYPE_SIZE (type)) == INTEGER_CST)
 	{
+	  tree field;
+	  /* A record which has any BLKmode members must itself be BLKmode;
+	     it can't go in a register.  */
+	  for (field = TYPE_FIELDS (type); field; field = TREE_CHAIN (field))
+	    if (TYPE_MODE (TREE_TYPE (field)) == BLKmode)
+	      goto record_lose;
+
 	  TYPE_MODE (type) 
 	    = agg_mode (TREE_INT_CST_LOW (TYPE_SIZE (type))
 			* TYPE_SIZE_UNIT (type));
+	record_lose: ;
 	}
       break;
 
@@ -656,9 +754,17 @@ layout_type (type)
       TYPE_MODE (type) = BLKmode;
       if (TREE_CODE (TYPE_SIZE (type)) == INTEGER_CST)
 	{
+	  tree field;
+	  /* A union which has any BLKmode members must itself be BLKmode;
+	     it can't go in a register.  */
+	  for (field = TYPE_FIELDS (type); field; field = TREE_CHAIN (field))
+	    if (TYPE_MODE (TREE_TYPE (field)) == BLKmode)
+	      goto union_lose;
+
 	  TYPE_MODE (type) 
 	    = agg_mode (TREE_INT_CST_LOW (TYPE_SIZE (type))
 			* TYPE_SIZE_UNIT (type));
+	union_lose: ;
 	}
       break;
 
@@ -669,10 +775,27 @@ layout_type (type)
       TYPE_ALIGN (type) = POINTER_BOUNDARY;
       break;
 
+    case METHOD_TYPE:
+      {
+	tree t = TREE_TYPE (type);
+	layout_type (t);
+	TYPE_MODE (type) = TYPE_MODE (t);
+	TYPE_SIZE (type) = TYPE_SIZE (t);
+	TYPE_SIZE_UNIT (type) = TYPE_SIZE_UNIT (t);
+	TYPE_ALIGN (type) = POINTER_BOUNDARY;
+      }
+      break;
+
     default:
       abort ();
     } /* end switch */
 
+  /* Evaluate nonconstant size only once, either now or as soon as safe.  */
+  if (TYPE_SIZE (type) != 0 && ! TREE_LITERAL (TYPE_SIZE (type)))
+    TYPE_SIZE (type) = variable_size (TYPE_SIZE (type));
+	
+  if (temporary)
+    resume_temporary_allocation ();
   resume_momentary (old);
 }
 
