@@ -25,9 +25,18 @@ and this notice must be preserved on all copies.  */
 #include "rtl.h"
 #include "gdbfiles.h"
 #include <stdio.h>
+#undef NULL
+#include "stddef.h"
+
+/* Get N_SO from stab.h if we can expect the file to exist.  */
+#ifdef DBX_DEBUGGING_INFO
+#include <stab.h>
+#endif
 
 /* .stabs code for source file name.  */
+#ifndef N_SO
 #define	N_SO 0x64
+#endif
 
 /* Unix maximum on file name length.  Needed for getwd.  */
 #define MAXNAMLEN 1024
@@ -66,6 +75,12 @@ static struct typevec_elt *typevec;
 
 static int total_types;
 
+/* Lists of types to which forward references have been made.
+   Separate lists for temporary and permanent types.  */
+
+static tree temporary_fwd_refs;
+static tree permanent_fwd_refs;
+
 /* `blockvec' is a chain recording all the symbol-blocks that have been output,
    giving the address-in-the-symseg of each one.  */
 
@@ -113,7 +128,7 @@ symout_init (filename, asm_file, sourcename)
 	   "Ltext:\t.stabs \"%s\",%d,0,0,Ltext\n",
 	   sourcename, N_SO);
   fprintf (asmfile, ".data 0\nLdata:\n");
-  ASM_OUTPUT_LOCAL (asmfile, "Lbss", 0);
+  ASM_OUTPUT_LOCAL (asmfile, "Lbss", 0, 0);
   fprintf (asmfile, ".gdbsym Ldata,%d\n",
 	   (char *) &buffer.databeg - (char *) &buffer);
   fprintf (asmfile, ".gdbsym Lbss,%d\n",
@@ -130,6 +145,9 @@ symout_init (filename, asm_file, sourcename)
   total_types = 0;
   total_blocks = 0;
 
+  permanent_fwd_refs = 0;
+  temporary_fwd_refs = 0;
+
   bzero (&buffer, sizeof buffer);
   fwrite (&buffer, sizeof buffer, 1, symfile);
 
@@ -240,6 +258,9 @@ symout_types (types)
   register struct typerec *records;
   register tree next;
   struct type buffer;
+  int this_run_address = next_address;
+
+  /* Count the number of types to be handled here.  */
 
   for (next = types, n_types = 0;
        next;
@@ -247,6 +268,9 @@ symout_types (types)
 
   records = (struct typerec *) alloca (n_types * sizeof (struct typerec));
 
+  /* Compute the amount of space each type needs, updating next_address
+     and storing the address of the data for each type.  */
+
   for (next = types, i = 0;
        next;
        next = TREE_CHAIN (next), i++)
@@ -296,6 +320,24 @@ symout_types (types)
 	  records[i].name_prefix = 0;
 	}
 
+      /* If this type was forward-referenced from a previous call
+	 to symout_types, store this type's address into the reference.  */
+      if (TYPE_POINTER_TO (next) != 0
+	  && TYPE_SYMTAB_ADDRESS (TYPE_POINTER_TO (next)) != 0
+	  && TYPE_SYMTAB_ADDRESS (TYPE_POINTER_TO (next)) < this_run_address)
+	{
+	  int pos = ftell (symfile);
+	  int myaddr = next_address;
+	  fflush (symfile);
+	  fseek (symfile,
+		 (TYPE_SYMTAB_ADDRESS (TYPE_POINTER_TO (next))
+		  + offsetof (struct type, target_type)),
+		 0);
+	  fwrite (&myaddr, sizeof (int), 1, symfile);
+	  fflush (symfile);
+	  fseek (symfile, pos, 0);
+	}
+
       records[i].address = next_address;
       TYPE_SYMTAB_ADDRESS (next) = next_address;
       velt->address = next_address;
@@ -326,6 +368,9 @@ symout_types (types)
       next_address += records[i].nfields * sizeof (struct field);
     }
 
+  /* Now write the data whose space we have assigned.
+     First fill the data into BUFFER, then write BUFFER.  */
+
   for (next = types, i = 0;
        next;
        next = TREE_CHAIN (next), i++)
@@ -334,12 +379,25 @@ symout_types (types)
 	symout_strings_print (records[i].name_prefix, 0,
 			      records[i].name, 0);
 
+      if (TREE_TYPE (next) != 0 && TYPE_OUTPUT_ADDRESS (TREE_TYPE (next)) == 0)
+	{
+	  /* We are making a forward-reference to our target type.
+	     Make a list of all of these.  */
+	  if (TREE_PERMANENT (next))
+	    permanent_fwd_refs
+	      = perm_tree_cons (TREE_TYPE (next), 0, permanent_fwd_refs);
+	  else
+	    temporary_fwd_refs
+	      = tree_cons (TREE_TYPE (next), 0, temporary_fwd_refs);
+	}
+
       if (TYPE_SIZE (next) == 0)
 	buffer.length = 0;
       else
 	buffer.length
 	  = (TREE_INT_CST_LOW (TYPE_SIZE (next))
 	     * TYPE_SIZE_UNIT (next) / BITS_PER_UNIT);
+
       buffer.name = (char *) records[i].name_address;
       buffer.target_type = (struct type *) (TREE_TYPE (next) ? TYPE_OUTPUT_ADDRESS (TREE_TYPE (next)) : 0);
 
@@ -401,6 +459,10 @@ symout_types (types)
 
       fwrite (&buffer, sizeof buffer, 1, symfile);
 
+      /* Now write the `struct field's that certain kinds of type have.
+	 This allocates space for the names of those fields,
+	 incrementing next_address for the names.  */
+
       switch (TREE_CODE (next))
 	{
 	case ARRAY_TYPE:
@@ -423,6 +485,9 @@ symout_types (types)
 	}
     }
 
+  /* Now output the strings referred to by the fields of certain types.
+     (next_address was already updated for these strings.)  */
+
   for (next = types, i = 0;
        next;
        next = TREE_CHAIN (next), i++)
@@ -440,6 +505,26 @@ symout_types (types)
 	}
     }
 }
+
+/* Given a list of types TYPES, return a chain of just those
+   that haven't been written in the symbol table.  */
+
+static tree
+filter_undefined_types (types)
+     tree types;
+{
+  tree new = 0;
+  tree next;
+
+  for (next = types; next; next = TREE_CHAIN (next))
+    if (TYPE_SYMTAB_ADDRESS (TREE_PURPOSE (next)) == 0)
+      {
+	TREE_CHAIN (TREE_PURPOSE (next)) = new;
+	new = TREE_PURPOSE (next);
+      }
+
+  return new;
+}
 
 /* Return nonzero if TYPE's range of possible values
    is not the full range allowed by the number of bits it has.
@@ -467,14 +552,23 @@ subrange_p (type)
     }
 
   if (uns)
-    return integer_zerop (TYPE_MIN_VALUE (type))
-      && (TREE_INT_CST_LOW (TYPE_MAX_VALUE (type))
-	  == (1 << TYPE_PRECISION (type)) - 1);
+    {
+      int mask;
+
+      if (TYPE_PRECISION (type) == HOST_BITS_PER_INT)
+	/* Shifting by 32 loses on some machines.  */
+	mask = -1;
+      else
+	mask = (1 << TYPE_PRECISION (type)) - 1;
+
+      return (integer_zerop (TYPE_MIN_VALUE (type))
+	      && (TREE_INT_CST_LOW (TYPE_MAX_VALUE (type)) == mask));
+    }
   else
-    return (TREE_INT_CST_LOW (TYPE_MIN_VALUE (type))
-	    == (-1) << (TYPE_PRECISION (type) - 1))
-      && (TREE_INT_CST_LOW (TYPE_MAX_VALUE (type))
-	  == (1 << (TYPE_PRECISION (type) - 1)) - 1);
+    return ((TREE_INT_CST_LOW (TYPE_MIN_VALUE (type))
+	     == (-1) << (TYPE_PRECISION (type) - 1))
+	    && (TREE_INT_CST_LOW (TYPE_MAX_VALUE (type))
+		== (1 << (TYPE_PRECISION (type) - 1)) - 1));
 }
 
 /* Functions to output the "fields" of various kinds of types.
@@ -855,6 +949,13 @@ symout_function (stmt, args, superblock_address)
     }
   return address;
 }
+
+symout_function_end ()
+{
+  /* Output dummy entries for any undefined structure references.  */
+  symout_types (filter_undefined_types (temporary_fwd_refs));
+  temporary_fwd_refs = 0;
+}
 
 /* Output all the data structure for a top two blocks in a compilation.
    The top block is for public (global) symbols;
@@ -1027,12 +1128,17 @@ symout_finish (filename, filetime)
      int filetime;
 {
   int *blockvector = (int *) alloca ((total_blocks + 1) * sizeof (int));
-  int *typevector = (int *) alloca ((total_types + 1) * sizeof (int));
+  int *typevector;
   int now = time (0);
   register int i;
   struct symbol_root buffer;
   char dir[MAXNAMLEN];
 
+  /* Output dummy entries for any undefined structure references.  */
+  symout_types (filter_undefined_types (permanent_fwd_refs));
+
+  typevector = (int *) alloca ((total_types + 1) * sizeof (int));
+
   buffer.language = language_c;
   buffer.blockvector = (struct blockvector *) next_address;
 
