@@ -39,17 +39,23 @@ and this notice must be preserved on all copies.  */
 
 #define	TARGET_DEFAULT 7
 
-/* Names to predefine in the preprocessor for this target machine.  */
+/* Define __HAVE_FPU__ in preprocessor, unless -msoft-float is specified.
+   This will control the use of inline 68881 insns in certain macros.  */
 
 #ifdef HPUX_ASM
-/* This is needed because some programs, particularly GDB, need to
+/* __HPUX_ASM__ is needed because some programs, particularly GDB, need to
    know which assembler is being used so that the correct `asm'
    instructions can be used. */
-#define CPP_PREDEFINES "-Dhp9000s200 -Dhp9000s300 -DPWB -Dmc68k -Dmc68000 -Dhpux -Dunix -D__HPUX_ASM__"
+#define CPP_SPEC "%{!msoft-float:-D__HAVE_FPU__} -D__HPUX_ASM__"
 #else
-#define CPP_PREDEFINES "-Dhp9000s200 -Dhp9000s300 -DPWB -Dmc68k -Dmc68000 -Dhpux -Dunix"
+#define CPP_SPEC "%{!msoft-float:-D__HAVE_FPU__}"
 #endif
 
+/* Names to predefine in the preprocessor for this target machine
+   (for non-strict-ANSI programs only).  */
+
+#define CPP_PREDEFINES "-Dhp9000s200 -Dhp9000s300 -DPWB -Dmc68k -Dmc68000 -Dhpux -Dunix"
+
 /* Every structure or union's size must be a multiple of 2 bytes.  */
 
 #define STRUCTURE_SIZE_BOUNDARY 16
@@ -99,8 +105,6 @@ and this notice must be preserved on all copies.  */
 
 #define TARGET_VERSION printf (" (68k, SGS/hpux syntax)");
 
-#define NO_DBX_FORMAT
-
 #define ASM_SPEC "%{m68000:+X}"
 
 #define REGISTER_NAMES \
@@ -135,7 +139,7 @@ and this notice must be preserved on all copies.  */
   else if (mask) fprintf (FILE, "\tmovm.l &0x%x,-(%%sp)\n", mask); }
 
 #define FUNCTION_PROFILER(FILE, LABEL_NO) \
-   fprintf (FILE, "\tmov.l &LP%d,%%d0\n\tjsr mcount\n", (LABEL_NO));
+   fprintf (FILE, "\tmov.l &LP%d,%%a0\n\tjsr mcount\n", (LABEL_NO));
 
 #define FUNCTION_EPILOGUE(FILE, SIZE) \
 { register int regno;						\
@@ -196,33 +200,33 @@ and this notice must be preserved on all copies.  */
     fprintf (FILE, "\trtd &%d\n", current_function_args_size);	\
   else fprintf (FILE, "\trts\n"); }
 
-#define ASM_FILE_START ""
+#define ASM_FILE_START(FILE)
 
 #define ASM_APP_ON ""
 
 #define ASM_APP_OFF ""
 
-#define TEXT_SECTION_ASM_OP "\ttext"
+#define TEXT_SECTION_ASM_OP "text"
 
-#define DATA_SECTION_ASM_OP "\tdata"
+#define DATA_SECTION_ASM_OP "data"
 
 #define	ASCII_DATA_ASM_OP "\tbyte"
 
 /* This says how to output an assembler line
    to define a global common symbol.  */
 
-#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs ("\tcomm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d\n", (SIZE)))
+  fprintf ((FILE), ",%d\n", (ROUNDED)))
 
 /* This says how to output an assembler line
    to define a local common symbol.  */
 
-#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs ("\tlcomm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d,2\n", (SIZE)))
+  fprintf ((FILE), ",%d,2\n", (ROUNDED)))
 
 /* Store in OUTPUT a string (made with alloca) containing
    an assembler-name for a local static variable named NAME.
@@ -245,22 +249,6 @@ do{  if (PREFIX[0] == 'L' && PREFIX[1] == 'I')		\
     fprintf (FILE, "%s%d:\n", PREFIX, NUM);		\
 } while(0)
 
-#if 0
-/* Are these definitions necessary? */
-
-#define ASM_OUTPUT_DOUBLE(FILE,VALUE)  \
-do { union { double d; long l[2]; } tem;		\
-     tem.d = (VALUE);					\
-     fprintf(FILE, "\tlong 0x%x,0x%x\n", tem.l[0], tem.l[1]); \
-   } while (0)
-
-#define ASM_OUTPUT_FLOAT(FILE,VALUE)  \
-do { union { float f; long l;} tem;			\
-     tem.f = (VALUE);					\
-     fprintf (FILE, "\tlong 0x%x\n", tem.l);	\
-   } while (0)
-#endif
-
 #define ASM_OUTPUT_DOUBLE(FILE, VALUE)					\
   fprintf (FILE, "\tdouble 0f%.20g\n", (VALUE))
 
