@@ -266,11 +266,14 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 
    I : Matches integers which are valid shift amounts for scaled indexing.
        These are 0, 1, 2, 3 for byte, word, double, and quadword.
-   J : Matches integers which fit a "quick" operand.   */
+   J : Matches integers which fit a "quick" operand.
+   K : Matches integers 0 to 7 (for inss and exts instructions).  */
 
 #define CONST_OK_FOR_LETTER_P(VALUE, C)  \
-  ((C) == 'I' ? (0 <= (VALUE) && (VALUE) <= 3) : \
-   (C) == 'J' ? (-8 <= (VALUE) && (VALUE) <= 7) : 0)
+  ((VALUE) < 8 && (VALUE) + 8 >= 0 ?		\
+   ((C) == 'I' ? (0 <= (VALUE) && (VALUE) <= 3) : \
+    (C) == 'J' ? (VALUE) <= 7 :			\
+    (C) == 'K' ? 0 <= (VALUE) : 0) : 0)
 
 /* Similar, but for floating constants, and defining letters G and H.
    Here VALUE is the CONST_DOUBLE rtx itself.  */
@@ -314,7 +317,7 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 #define PUSH_ROUNDING(BYTES) (BYTES)
 
 /* Offset of first parameter from the argument pointer register value.  */
-#define FIRST_PARM_OFFSET 8
+#define FIRST_PARM_OFFSET(FNDECL) 8
 
 /* Value is 1 if returning from a function call automatically
    pops the arguments described by the number-of-args field in the call.
@@ -429,6 +432,10 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 		      : GET_MODE_SIZE (MODE))))  		\
  ? 2 - (CUM) / 4 : 0)
 
+#ifndef MAIN_FUNCTION_PROLOGUE
+#define MAIN_FUNCTION_PROLOGUE
+#endif
+
 /* This macro generates the assembly code for function entry.
    FILE is a stdio stream to output the code to.
    SIZE is an int: how many units of temporary storage to allocate.
@@ -443,6 +450,7 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
   char used_regs_buf[32], *bufp = used_regs_buf;		\
   int used_fregs_buf[8], *fbufp = used_fregs_buf;		\
   extern char call_used_regs[];					\
+  MAIN_FUNCTION_PROLOGUE;					\
   for (regno = 0, nregs = 0; regno < 8; regno++)		\
     if (regs_ever_live[regno] && !call_used_regs[regno]) {	\
        nregs += 1;						\
@@ -726,8 +734,6 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
  from registers which are not FP or SP (or SB or PC).  This
  makes _x(fp) valid, while _x(r0) is invalid.  */
 
-# define SEQUENT_HAS_FIXED_THEIR_BUG 0
-
 #define INDIRECTABLE_1_ADDRESS_P(X)  \
   (CONSTANT_P (X)							\
    || (GET_CODE (X) == REG && REG_OK_FOR_BASE_P (X))			\
@@ -801,18 +807,18 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
       else if (GET_CODE (XEXP (xfooy, 0)) == CONST_INT			\
 	  && GET_CODE (XEXP (xfooy, 1)) == PLUS)			\
 	xfooy = XEXP (xfooy, 1);					\
-	xfooz = XEXP (xfooy, 1);					\
-        if (INDEX_TERM_P (xfooz, MODE))					\
-	  { rtx t = XEXP (xfooy, 0); GO_IF_NONINDEXED_ADDRESS (t, ADDR); } \
-        xfooz = XEXP (xfooy, 0);					\
-	if (INDEX_TERM_P (xfooz, MODE))					\
-	  { rtx t = XEXP (xfooy, 1); GO_IF_NONINDEXED_ADDRESS (t, ADDR); } \
+      xfooz = XEXP (xfooy, 1);						\
+      if (INDEX_TERM_P (xfooz, MODE))					\
+	{ rtx t = XEXP (xfooy, 0); GO_IF_NONINDEXED_ADDRESS (t, ADDR); }\
+      xfooz = XEXP (xfooy, 0);						\
+      if (INDEX_TERM_P (xfooz, MODE))					\
+	{ rtx t = XEXP (xfooy, 1); GO_IF_NONINDEXED_ADDRESS (t, ADDR); }\
     }									\
   else if (INDEX_TERM_P (xfooy, MODE))					\
     goto ADDR;								\
   else if (GET_CODE (xfooy) == PRE_DEC)					\
     if (REGNO (XEXP (xfooy, 0)) == STACK_POINTER_REGNUM) goto ADDR;	\
-    else abort ();							\
+  else abort ();							\
 }
 
 /* Try machine-dependent ways of modifying an illegitimate address
@@ -985,7 +991,7 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 
 /* Output at beginning of assembler file.  */
 
-#define ASM_FILE_START "#NO_APP\n"
+#define ASM_FILE_START(FILE) fprintf (FILE, "#NO_APP\n");
 
 /* Output to assembler file text saying following lines
    may contain character constants, extra white space, comments, etc.  */
@@ -1109,18 +1115,18 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 /* This says how to output an assembler line
    to define a global common symbol.  */
 
-#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs (".comm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d\n", (SIZE)))
+  fprintf ((FILE), ",%d\n", (ROUNDED)))
 
 /* This says how to output an assembler line
    to define a local common symbol.  */
 
-#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs (".lcomm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d\n", (SIZE)))
+  fprintf ((FILE), ",%d\n", (ROUNDED)))
 
 /* Store in OUTPUT a string (made with alloca) containing
    an assembler-name for a local static variable named NAME.
@@ -1153,6 +1159,7 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 
 #define PRINT_OPERAND(FILE, X, CODE)  \
 { if (CODE == '$') fprintf (FILE, "$");					\
+  else if (CODE == '?');						\
   else if (GET_CODE (X) == REG)						\
     fprintf (FILE, "%s", reg_name [REGNO (X)]);				\
   else if (GET_CODE (X) == MEM)						\
@@ -1162,7 +1169,10 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
       { union { double d; int i[2]; } u;				\
 	u.i[0] = XINT (X, 0); u.i[1] = XINT (X, 1);			\
 	fprintf (FILE, "$0d%.20e", u.d); }				\
-    else fprintf (FILE, "$0f%.20e", XINT (X, 0));			\
+    else								\
+      { union { double d; int i[2]; } u;				\
+	u.i[0] = XINT (X, 0); u.i[1] = XINT (X, 1);			\
+	fprintf (FILE, "$0f%.20e", u.d); }				\
   else { putc ('$', FILE); output_addr_const (FILE, X); }}
 
 /* Print a memory operand whose address is X, on file FILE.  */
@@ -1298,18 +1308,21 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 		"XbwXdXXXq"[INTVAL (XEXP (reg1, 1))]);		        \
 	      reg1 = 0;							\
 	    }								\
-	  else if (GET_CODE (XEXP (addr, 0)) == REG)			\
+	  else if (GET_CODE (XEXP (addr, 0)) == REG			\
+		   && REGNO (XEXP (addr, 0)) < 8)			\
 	    {								\
 	      sprintf (reg2_str, "[%s:b]",				\
 		reg_name[ REGNO(XEXP (addr, 0)) ]);			\
 	      addr = XEXP (addr, 1); /* CONST / REG */			\
 	    }								\
-	  else if (GET_CODE (XEXP (addr, 1)) == REG)			\
+	  else if (GET_CODE (XEXP (addr, 1)) == REG			\
+		   && REGNO (XEXP (addr, 1)) < 8)			\
 	    {								\
 	      sprintf (reg2_str, "[%s:b]",				\
 		reg_name[ REGNO(XEXP (addr, 1)) ]);			\
 	      addr = XEXP (addr, 0); /* CONST / REG */			\
 	    }								\
+	  else abort ();						\
       }									\
       if (addr)								\
 	switch (GET_CODE (addr)) {					\
@@ -1341,9 +1354,42 @@ enum reg_class { NO_REGS, GENERAL_REGS, FLOAT_REGS, GEN_AND_FLOAT_REGS,
 		} else if (!*reg2_str)					\
 		    sprintf (reg2_str, "[%s:b]",			\
 			reg_name[REGNO (addr)]);			\
-		else {							\
+		else abort();						\
+		break;							\
+	    case MEM:							\
+		addr = XEXP(addr,0);					\
+		switch (GET_CODE(addr)) {				\
+		case REG:						\
+		  if (!*reg1_str) {					\
+		    if (offset || offset_printed)			\
+		      sprintf (reg1_str, "(0(%s))",			\
+			       reg_name[REGNO (addr)]); 		\
+		    else						\
+		      sprintf (reg1_str, "0(0(%s))",			\
+			       reg_name[REGNO (addr)]);			\
+		  } else						\
 		    abort();						\
-	        }							\
+		  break;						\
+		case PLUS:						\
+		  if (GET_CODE (XEXP (addr, 0)) == REG) {		\
+		    if (!*reg1_str) {					\
+		      sprintf (reg1_str, "(%s))",			\
+			       reg_name[REGNO(XEXP(addr, 0))]);		\
+		      offset = XEXP(addr, 1);				\
+		    } else						\
+		      abort();						\
+		  } else {						\
+		    if (!*reg1_str) {					\
+		      sprintf (reg1_str, "(%s))",			\
+			       reg_name[REGNO(XEXP(addr, 1))]);		\
+		      offset = XEXP(addr, 0);				\
+		    } else						\
+		      abort();						\
+		  }							\
+		  break;						\
+		default:						\
+		  abort();						\
+		}							\
 		break;							\
 	    default:							\
 		if (offset_printed)					\
