@@ -60,7 +60,9 @@ extern int target_flags;
 
 /* Default target_flags if no switches specified.  */
 
+#ifndef TARGET_DEFAULT
 #define TARGET_DEFAULT 1
+#endif
 
 /* Target machine storage layout */
 
@@ -108,6 +110,9 @@ extern int target_flags;
 /* Every structure's size must be a multiple of this.  */
 #define STRUCTURE_SIZE_BOUNDARY 8
 
+/* A bitfield declared as `int' forces `int' alignment for the struct.  */
+#define PCC_BITFIELD_TYPE_MATTERS
+
 /* No data type wants to be aligned rounder than this.  */
 #define BIGGEST_ALIGNMENT (TARGET_VAXC_ALIGNMENT ? 8 : 32)
 
@@ -293,7 +298,7 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
 #define PUSH_ROUNDING(BYTES) (BYTES)
 
 /* Offset of first parameter from the argument pointer register value.  */
-#define FIRST_PARM_OFFSET 4
+#define FIRST_PARM_OFFSET(FNDECL) 4
 
 /* Value is 1 if returning from a function call automatically
    pops the arguments described by the number-of-args field in the call.
@@ -694,6 +699,45 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
 
 #define UDIVSI3_LIBCALL "*udiv"
 #define UMODSI3_LIBCALL "*urem"
+
+/* Check a `double' value for validity for a particular machine mode.  */
+
+/* note that it is very hard to accidently create a number that fits in a
+   double but not in a float, since their ranges are almost the same */
+#define CHECK_FLOAT_VALUE(mode, d) \
+  if ((mode) == SFmode) \
+    { \
+      if ((d) > 1.7014117331926443e+38) \
+	{ error ("magnitude of constant too large for `float'"); \
+	  (d) = 1.7014117331926443e+38; } \
+      else if ((d) < -1.7014117331926443e+38) \
+	{ error ("magnitude of constant too large for `float'"); \
+	  (d) = -1.7014117331926443e+38; } \
+      else if (((d) > 0) && ((d) < 2.9387358770557188e-39)) \
+	{ warning ("`float' constant truncated to zero"); \
+	  (d) = 0.0; } \
+      else if (((d) < 0) && ((d) > -2.9387358770557188e-39)) \
+	{ warning ("`float' constant truncated to zero"); \
+	  (d) = 0.0; } \
+    }
+
+/* For future reference:
+   D Float: 9 bit, sign magnitude, excess 128 binary exponent
+            normalized 56 bit fraction, redundant bit not represented
+            approximately 16 decimal digits of precision
+
+   The values to use if we trust decimal to binary conversions:
+#define MAX_D_FLOAT 1.7014118346046923e+38
+#define MIN_D_FLOAT .29387358770557188e-38
+
+   G float: 12 bit, sign magnitude, excess 1024 binary exponent
+            normalized 53 bit fraction, redundant bit not represented
+            approximately 15 decimal digits precision
+
+   The values to use if we trust decimal to binary conversions:
+#define MAX_G_FLOAT .898846567431157e+308
+#define MIN_G_FLOAT .556268464626800e-308
+*/
 
 /* Tell final.c how to eliminate redundant test instructions.  */
 
@@ -740,7 +784,7 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
 
 /* Output at beginning of assembler file.  */
 
-#define ASM_FILE_START "#NO_APP\n"
+#define ASM_FILE_START(FILE) fprintf (FILE, "#NO_APP\n");
 
 /* Output to assembler file text saying following lines
    may contain character constants, extra white space, comments, etc.  */
@@ -767,6 +811,10 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
 {"r0", "r1", "r2", "r3", "r4", "r5", "r6", "r7", "r8", \
  "r9", "r10", "r11", "ap", "fp", "sp", "pc"}
 
+/* This is BSD, so it wants DBX format.  */
+
+#define DBX_DEBUGGING_INFO
+
 /* How to renumber registers for dbx and gdb.
    Vax needs no change in the numeration.  */
 
@@ -876,24 +924,27 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
 #define ASM_OUTPUT_ALIGN(FILE,LOG)  \
   fprintf (FILE, "\t.align %d\n", (LOG))
 
+/* This is how to output an assembler line
+   that says to advance the location counter by SIZE bytes.  */
+
 #define ASM_OUTPUT_SKIP(FILE,SIZE)  \
   fprintf (FILE, "\t.space %d\n", (SIZE))
 
 /* This says how to output an assembler line
    to define a global common symbol.  */
 
-#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_COMMON(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs (".comm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d\n", (SIZE)))
+  fprintf ((FILE), ",%d\n", (ROUNDED)))
 
 /* This says how to output an assembler line
    to define a local common symbol.  */
 
-#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE)  \
+#define ASM_OUTPUT_LOCAL(FILE, NAME, SIZE, ROUNDED)  \
 ( fputs (".lcomm ", (FILE)),			\
   assemble_name ((FILE), (NAME)),		\
-  fprintf ((FILE), ",%d\n", (SIZE)))
+  fprintf ((FILE), ",%d\n", (ROUNDED)))
 
 /* Store in OUTPUT a string (made with alloca) containing
    an assembler-name for a local static variable named NAME.
@@ -937,86 +988,8 @@ enum reg_class { NO_REGS, ALL_REGS, LIM_REG_CLASSES };
       fprintf (FILE, "$0%c%.20e", ASM_DOUBLE_CHAR, u.d); }		\
   else { putc ('$', FILE); output_addr_const (FILE, X); }}
 
-/* Print a memory operand whose address is X, on file FILE.  */
+/* Print a memory operand whose address is X, on file FILE.
+   This uses a function in output-vax.c.  */
 
 #define PRINT_OPERAND_ADDRESS(FILE, ADDR)  \
-{ register rtx reg1, reg2, breg, ireg;					\
-  register rtx addr = ADDR;						\
-  rtx offset;								\
- retry:									\
-  switch (GET_CODE (addr))						\
-    {									\
-    case MEM:								\
-      fprintf (FILE, "*");						\
-      addr = XEXP (addr, 0);						\
-      goto retry;							\
-    case REG:								\
-      fprintf (FILE, "(%s)", reg_name [REGNO (addr)]);			\
-      break;								\
-    case PRE_DEC:							\
-      fprintf (FILE, "-(%s)", reg_name [REGNO (XEXP (addr, 0))]);	\
-      break;								\
-    case POST_INC:							\
-      fprintf (FILE, "(%s)+", reg_name [REGNO (XEXP (addr, 0))]);	\
-      break;								\
-    case PLUS:								\
-      reg1 = 0;	reg2 = 0;						\
-      ireg = 0;	breg = 0;						\
-      offset = 0;							\
-      if (CONSTANT_ADDRESS_P (XEXP (addr, 0))				\
-	  || GET_CODE (XEXP (addr, 0)) == MEM)				\
-	{								\
-	  offset = XEXP (addr, 0);					\
-	  addr = XEXP (addr, 1);					\
-	}								\
-      else if (CONSTANT_ADDRESS_P (XEXP (addr, 1))			\
-	       || GET_CODE (XEXP (addr, 1)) == MEM)			\
-	{								\
-	  offset = XEXP (addr, 1);					\
-	  addr = XEXP (addr, 0);					\
-	}								\
-      if (GET_CODE (addr) != PLUS) ;					\
-      else if (GET_CODE (XEXP (addr, 0)) == MULT)			\
-	{								\
-	  reg1 = XEXP (addr, 0);					\
-	  addr = XEXP (addr, 1);					\
-	}								\
-      else if (GET_CODE (XEXP (addr, 1)) == MULT)			\
-	{								\
-	  reg1 = XEXP (addr, 1);					\
-	  addr = XEXP (addr, 0);					\
-	}								\
-      else if (GET_CODE (XEXP (addr, 0)) == REG)			\
-	{								\
-	  reg1 = XEXP (addr, 0);					\
-	  addr = XEXP (addr, 1);					\
-	}								\
-      else if (GET_CODE (XEXP (addr, 1)) == REG)			\
-	{								\
-	  reg1 = XEXP (addr, 1);					\
-	  addr = XEXP (addr, 0);					\
-	}								\
-      if (GET_CODE (addr) == REG || GET_CODE (addr) == MULT)		\
-	{ if (reg1 == 0) reg1 = addr; else reg2 = addr; addr = 0; }	\
-      if (offset != 0) { if (addr != 0) abort (); addr = offset; }	\
-      if (reg1 != 0 && GET_CODE (reg1) == MULT)				\
-	{ breg = reg2; ireg = reg1; }					\
-      else if (reg2 != 0 && GET_CODE (reg2) == MULT)			\
-	{ breg = reg1; ireg = reg2; }					\
-      else if (reg2 != 0 || GET_CODE (addr) == MEM)			\
-	{ breg = reg2; ireg = reg1; }					\
-      else								\
-	{ breg = reg1; ireg = reg2; }					\
-      if (addr != 0)							\
-	output_address (offset);					\
-      if (breg != 0)							\
-	{ if (GET_CODE (breg) != REG) abort ();				\
-	  fprintf (FILE, "(%s)", reg_name[REGNO (breg)]); }		\
-      if (ireg != 0)							\
-	{ if (GET_CODE (ireg) == MULT) ireg = XEXP (ireg, 0);		\
-	  if (GET_CODE (ireg) != REG) abort ();				\
-	  fprintf (FILE, "[%s]", reg_name[REGNO (ireg)]); }		\
-      break;								\
-    default:								\
-      output_addr_const (FILE, addr);					\
-    }}
+ print_operand_address (FILE, ADDR)
