@@ -28,6 +28,9 @@ and this notice must be preserved on all copies.  */
 #include <stdio.h>
 #include <signal.h>
 
+#include <sys/types.h>
+#include <sys/stat.h>
+
 #ifdef USG
 #include <sys/param.h>
 #include <sys/times.h>
@@ -39,11 +42,6 @@ and this notice must be preserved on all copies.  */
 #endif
 #endif
 
-#ifndef _TYPES_
-#include <sys/types.h>
-#endif
-#include <sys/stat.h>
-
 #include "tree.h"
 #include "c-tree.h"
 #include "rtl.h"
@@ -79,7 +77,8 @@ int target_flags;
 char *input_filename;
 
 /* Name of top-level original source file (what was input to cpp).
-   This comes from the first #-command in the actual input.  */
+   This comes from the #-command at the beginning of the actual input.
+   If there isn't any there, then this is the cc1 input file name.  */
 
 char *main_input_filename;
 
@@ -109,9 +108,15 @@ int global_reg_dump = 0;
 int jump2_opt_dump = 0;
 
 /* 1 => write gdb debugging output (using symout.c).  -g
-   2 => write dbx debugging output (using dbxout.c).  -G  */
+   2 => write dbx debugging output (using dbxout.c).  -G
+   3 => write sdb debugging output (using sdbout.c).  -g.  */
 
-int write_symbols = 0;
+enum debugger write_symbols = NO_DEBUG;
+
+/* Nonzero means can use our own extensions to DBX format.
+   Relevant only with write_symbols == DBX_DEBUG.  */
+
+int use_gdb_dbx_extensions;
 
 /* Nonzero means do optimizations.  -opt.  */
 
@@ -163,32 +168,6 @@ int flag_omit_frame_pointer = 0;
 
 int flag_no_peephole = 0;
 
-/* Nonzero means `char' should be signed.  */
-
-int flag_signed_char;
-
-/* Nonzero means allow type mismatches in conditional expressions;
-   just make their values `void'.   */
-
-int flag_cond_mismatch;
-
-/* Nonzero means don't recognize the keyword `asm'.  */
-
-int flag_no_asm;
-
-/* Nonzero means warn about implicit declarations.  */
-
-int warn_implicit;
-
-/* Nonzero means warn about function definitions that default the return type
-   or that use a null return and have a return-type other than void.  */
-
-int warn_return_type;
-
-/* Nonzero means do some things the same way PCC does.  */
-
-int flag_traditional;
-
 /* Nonzero means all references through pointers are volatile.  */
 
 int flag_volatile;
@@ -216,13 +195,14 @@ int extra_warnings = 0;
 
 int errorcount = 0;
 int warningcount = 0;
+int sorrycount = 0;
 
 /* Nonzero if generating code to do profiling.  */
 
 int profile_flag = 0;
 
 /* Nonzero for -pedantic switch: warn about anything
-   that standard C forbids.  */
+   that standard spec forbids.  */
 
 int pedantic = 0;
 
@@ -365,12 +345,19 @@ fatal_io_error (name)
 }
 
 void
-fatal (s)
+fatal (s, v)
      char *s;
 {
-  error (s, 0);
+  error (s, v);
   exit (34);
 }
+
+static int need_error_newline;
+
+/* Function of last error message;
+   more generally, function such that if next error message is in it
+   then we don't have to mention the function name.  */
+static tree last_error_function = NULL;
 
 /* Called when the start of a function definition is parsed,
    this function prints on stderr the name of the function.  */
@@ -384,34 +371,30 @@ announce_function (decl)
       fprintf (stderr, " %s", IDENTIFIER_POINTER (DECL_NAME (decl)));
       fflush (stderr);
     }
+  need_error_newline = 1;
+  last_error_function = current_function_decl;
 }
-
+
 /* Prints out, if necessary, the name of the current function
    which caused an error.  Called from all error and warning functions.  */
 
 void
 report_error_function()
 {
-  static tree last_error_function = NULL;
+  if (need_error_newline)
+    {
+      fprintf (stderr, "\n");
+      need_error_newline = 0;
+    }
 
   if (last_error_function != current_function_decl)
     {
       if (current_function_decl == NULL)
-	{
-	  if (!quiet_flag)
-	    fprintf (stderr, "\n");
-	  fprintf (stderr, "At top level:\n");
-	}
+	fprintf (stderr, "At top level:\n");
       else
-	{
-	  if (!quiet_flag)
-	    /* We already know this info.  Don't print it twice.
-	       But make sure we are at the beginning of a line.  */
-	    fprintf (stderr, "\n");
-	  else
-	    fprintf (stderr, "In function %s:\n",
-		     IDENTIFIER_POINTER (DECL_NAME (current_function_decl)));
-	}
+	fprintf (stderr, "In function %s:\n",
+		 IDENTIFIER_POINTER (DECL_NAME (current_function_decl)));
+
       last_error_function = current_function_decl;
     }
 }
@@ -420,22 +403,24 @@ report_error_function()
    S and V are a string and an arg for `printf'.  */
 
 void
-error (s, v)
+error (s, v, v2)
      char *s;
      int v;			/* @@also used as pointer */
+     int v2;			/* @@also used as pointer */
 {
-  error_with_file_and_line (input_filename, lineno, s, v);
+  error_with_file_and_line (input_filename, lineno, s, v, v2);
 }
 
 /* Report an error at line LINE of file FILE.
    S and V are a string and an arg for `printf'.  */
 
 void
-error_with_file_and_line (file, line, s, v)
+error_with_file_and_line (file, line, s, v, v2)
      char *file;
      int line;
      char *s;
      int v;
+     int v2;
 {
   count_error (0);
 
@@ -445,7 +430,7 @@ error_with_file_and_line (file, line, s, v)
     fprintf (stderr, "%s:%d: ", file, line);
   else
     fprintf (stderr, "cc1: ");
-  fprintf (stderr, s, v);
+  fprintf (stderr, s, v, v2);
   fprintf (stderr, "\n");
 }
 
@@ -475,10 +460,11 @@ error_with_decl (decl, s)
    S and V are a string and an arg for `printf'.  */
 
 void
-warning_with_line (line, s, v)
+warning_with_line (line, s, v, v2)
      int line;
      char *s;
      int v;
+     int v2;
 {
   if (count_error (1) == 0)
     return;
@@ -491,7 +477,7 @@ warning_with_line (line, s, v)
     fprintf (stderr, "cc1: ");
 
   fprintf (stderr, "warning: ");
-  fprintf (stderr, s, v);
+  fprintf (stderr, s, v, v2);
   fprintf (stderr, "\n");
 }
 
@@ -499,11 +485,12 @@ warning_with_line (line, s, v)
    S and V are a string and an arg for `printf'.  */
 
 void
-warning (s, v)
+warning (s, v, v2)
      char *s;
      int v;			/* @@also used as pointer */
+     int v2;
 {
-  warning_with_line (lineno, s, v);
+  warning_with_line (lineno, s, v, v2);
 }
 
 /* Report a warning at the declaration DECL.
@@ -529,6 +516,43 @@ warning_with_decl (decl, s)
     fprintf (stderr, s, "((anonymous))");
   fprintf (stderr, "\n");
 }
+
+/* Apologize for not implementing some feature.
+   S, V, and V2 are a string and args for `printf'.  */
+
+void
+sorry (s, v, v2)
+     char *s;
+     int v, v2;
+{
+  sorrycount++;
+  if (input_filename)
+    fprintf (stderr, "%s:%d: ", input_filename, lineno);
+  else
+    fprintf (stderr, "cc1: ");
+
+  fprintf (stderr, "sorry, not implemented: ");
+  fprintf (stderr, s, v, v2);
+  fprintf (stderr, "\n");
+}
+
+/* Apologize for not implementing some feature, then quit.
+   S, V, and V2 are a string and args for `printf'.  */
+
+void
+really_sorry (s, v, v2)
+     char *s;
+     int v, v2;
+{
+  if (input_filename)
+    fprintf (stderr, "%s:%d: ", input_filename, lineno);
+  else
+    fprintf (stderr, "c++: ");
+
+  fprintf (stderr, "sorry, not implemented: ");
+  fprintf (stderr, s, v, v2);
+  fatal (" (fatal)\n");
+}
 
 /* When `malloc.c' is compiled with `rcheck' defined,
    it calls this function to report clobberage.  */
@@ -739,6 +763,8 @@ compile_file (name)
     strcpy (dumpname, dump_base_name);
     if (len > 2 && ! strcmp (".c", dumpname + len - 2))
       dumpname[len - 2] = 0;
+    else if (len > 2 && ! strcmp (".i", dumpname + len - 2))
+      dumpname[len - 2] = 0;
     else if (len > 3 && ! strcmp (".co", dumpname + len - 3))
       dumpname[len - 3] = 0;
     strcat (dumpname, ".s");
@@ -750,7 +776,6 @@ compile_file (name)
     asm_out_file = fopen (asm_file_name, "w");
     if (asm_out_file == 0)
       pfatal_with_name (asm_file_name ? asm_file_name : dumpname);
-    fprintf (asm_out_file, ASM_FILE_START);
   }
 
   input_filename = name;
@@ -758,28 +783,43 @@ compile_file (name)
   /* the beginning of the file is a new line; check for # */
   /* With luck, we discover the real source file's name from that
      and put it in input_filename.  */
-  check_newline ();
+  ungetc (check_newline (), finput);
+
+  /* If the input doesn't start with a #line, use the input name
+     as the official input file name.  */
+  if (main_input_filename == 0)
+    main_input_filename = name;
+
+  ASM_FILE_START (asm_out_file);
 
   /* If GDB symbol table desired, open the GDB symbol output file.  */
-  if (write_symbols == 1)
+  if (write_symbols == GDB_DEBUG)
     {
       register char *dumpname = (char *) xmalloc (dump_base_name_length + 6);
       int len = strlen (dump_base_name);
       strcpy (dumpname, dump_base_name);
       if (len > 2 && ! strcmp (".c", dumpname + len - 2))
 	dumpname[len - 2] = 0;
+      else if (len > 2 && ! strcmp (".i", dumpname + len - 2))
+	dumpname[len - 2] = 0;
       else if (len > 3 && ! strcmp (".co", dumpname + len - 3))
 	dumpname[len - 3] = 0;
       strcat (dumpname, ".sym");
       if (sym_file_name == 0)
 	sym_file_name = dumpname;
-      symout_init (sym_file_name, asm_out_file, input_filename);
+      symout_init (sym_file_name, asm_out_file, main_input_filename);
     }
 
   /* If dbx symbol table desired, initialize writing it
      and output the predefined types.  */
-  if (write_symbols == 2)
+#ifdef DBX_DEBUGGING_INFO
+  if (write_symbols == DBX_DEBUG)
     dbxout_init (asm_out_file, main_input_filename);
+#endif
+#ifdef SDB_DEBUGGING_INFO
+  if (write_symbols == SDB_DEBUG)
+    sdbout_init (asm_out_file, main_input_filename);
+#endif
 
   /* Initialize yet another pass.  */
 
@@ -821,15 +861,26 @@ compile_file (name)
   }
 
   /* Do dbx symbols */
-  if (write_symbols == 2)
+#ifdef DBX_DEBUGGING_INFO
+  if (write_symbols == DBX_DEBUG)
     TIMEVAR (symout_time,
 	     {
 	       dbxout_tags (gettags ());
 	       dbxout_types (get_permanent_types ());
 	     });
+#endif
+
+#ifdef SDB_DEBUGGING_INFO
+  if (write_symbols == SDB_DEBUG)
+    TIMEVAR (symout_time,
+	     {
+	       sdbout_tags (gettags ());
+	       sdbout_types (get_permanent_types ());
+	     });
+#endif
 
   /* Do gdb symbols */
-  if (write_symbols == 1)
+  if (write_symbols == GDB_DEBUG)
     TIMEVAR (symout_time,
 	     {
 	       struct stat statbuf;
@@ -921,14 +972,21 @@ rest_of_decl_compilation (decl, asmspec, top_level, at_end)
   if (TREE_STATIC (decl) || TREE_EXTERNAL (decl))
     TIMEVAR (varconst_time,
 	     {
-	       assemble_variable (decl, asmspec, top_level, write_symbols, at_end);
+	       assemble_variable (decl, asmspec, top_level,
+				  write_symbols, at_end);
 	     });
-  else if (write_symbols == 2 && TREE_CODE (decl) == TYPE_DECL)
+#ifdef DBX_DEBUGGING_INFO
+  else if (write_symbols == DBX_DEBUG && TREE_CODE (decl) == TYPE_DECL)
     TIMEVAR (varconst_time, dbxout_symbol (decl, 0));
+#endif
+#ifdef SDB_DEBUGGING_INFO
+  else if (write_symbols == SDB_DEBUG && TREE_CODE (decl) == TYPE_DECL)
+    TIMEVAR (varconst_time, sdbout_symbol (decl, 0));
+#endif
 
   if (top_level)
     {
-      if (write_symbols == 1)
+      if (write_symbols == GDB_DEBUG)
 	{
 	  TIMEVAR (symout_time,
 		   {
@@ -1245,7 +1303,7 @@ rest_of_compilation (decl)
 
   /* Write GDB symbols if requested */
 
-  if (write_symbols == 1)
+  if (write_symbols == GDB_DEBUG)
     {
       TIMEVAR (symout_time,
 	       {
@@ -1255,6 +1313,7 @@ rest_of_compilation (decl)
 		 DECL_BLOCK_SYMTAB_ADDRESS (decl)
 		   = symout_function (DECL_INITIAL (decl),
 				      DECL_ARGUMENTS (decl), 0);
+		 symout_function_end ();
 	       });
     }
   else
@@ -1262,11 +1321,18 @@ rest_of_compilation (decl)
 
   /* Write DBX symbols if requested */
 
-  if (write_symbols == 2)
+#ifdef DBX_DEBUGGING_INFO
+  if (write_symbols == DBX_DEBUG)
     TIMEVAR (symout_time, dbxout_function (decl));
+#endif
 
  exit_rest_of_compilation:
 
+  /* Clear out the real_constant_chain before some of the rtx's
+     it runs through become garbage.  */
+
+  clear_const_double_mem ();
+
   /* The parsing time is all the time spent in yyparse
      *except* what is spent in this function.  */
 
@@ -1302,7 +1368,10 @@ main (argc, argv, envp)
   /* Initialize whether `char' is signed.  */
   flag_signed_char = DEFAULT_SIGNED_CHAR;
 
-  /* Initialize reg-sets now so switches may override.  */
+  /* This is zeroed by -O.  */
+  obey_regdecls = 1;
+
+  /* Initialize register usage now so switches may override.  */
   init_reg_sets ();
 
   target_flags = 0;
@@ -1367,8 +1436,6 @@ main (argc, argv, envp)
 	    register char *p = &str[1];
 	    if (!strcmp (p, "float-store"))
 	      flag_float_store = 1;
-	    else if (!strcmp (p, "traditional"))
-	      flag_traditional = 1;
 	    else if (!strcmp (p, "volatile"))
 	      flag_volatile = 1;
 	    else if (!strcmp (p, "defer-pop"))
@@ -1379,26 +1446,18 @@ main (argc, argv, envp)
 	      flag_omit_frame_pointer = 1;
 	    else if (!strcmp (p, "no-omit-frame-pointer"))
 	      flag_omit_frame_pointer = 0;
+	    else if (!strcmp (p, "writable-strings"))
+	      flag_writable_strings = 1;
 	    else if (!strcmp (p, "no-peephole"))
 	      flag_no_peephole = 1;
-	    else if (!strcmp (p, "signed-char"))
-	      flag_signed_char = 1;
-	    else if (!strcmp (p, "unsigned-char"))
-	      flag_signed_char = 0;
 	    else if (!strcmp (p, "force-mem"))
 	      flag_force_mem = 1;
 	    else if (!strcmp (p, "force-addr"))
 	      flag_force_addr = 1;
 	    else if (!strcmp (p, "combine-regs"))
 	      flag_combine_regs = 1;
-	    else if (!strcmp (p, "writable-strings"))
-	      flag_writable_strings = 1;
 	    else if (!strcmp (p, "no-function-cse"))
 	      flag_no_function_cse = 1;
-	    else if (!strcmp (p, "cond-mismatch"))
-	      flag_cond_mismatch = 1;
-	    else if (!strcmp (p, "no-asm"))
-	      flag_no_asm = 1;
 	    else if (!strncmp (p, "fixed-", 6))
 	      fix_register (&p[6], 1, 1);
 	    else if (!strncmp (p, "call-used-", 10))
@@ -1409,25 +1468,25 @@ main (argc, argv, envp)
 	      flag_inline_functions = 1;
 	    else if (!strcmp (p, "keep-inline-functions"))
 	      flag_keep_inline_functions = 1;
-	    else
-	      error ("Invalid option, `%s'", argv[i]);	      
+	    else if (! lang_decode_option (argv[i]))
+	      error ("Invalid option `%s'", argv[i]);	      
 	  }
 	else if (!strcmp (str, "noreg"))
-	  obey_regdecls = 1;
+	  ;
 	else if (!strcmp (str, "opt"))
-	  optimize = 1;
+	  optimize = 1, obey_regdecls = 0;
+	else if (!strcmp (str, "O"))
+	  optimize = 1, obey_regdecls = 0;
 	else if (!strcmp (str, "pedantic"))
 	  pedantic = 1;
-	else if (!strcmp (str, "traditional"))
-	  flag_traditional = 1;
-	else if (!strcmp (str, "ansi"))
-	  flag_no_asm = 1;
+	else if (lang_decode_option (argv[i]))
+	  ;
 	else if (!strcmp (str, "quiet"))
 	  quiet_flag = 1;
 	else if (!strcmp (str, "version"))
 	  {
-	    extern char *version_string;
-	    printf ("GNU C version %s", version_string);
+	    extern char *version_string, *language_string;
+	    printf ("%s version %s", language_string, version_string);
 #ifdef TARGET_VERSION
 	    TARGET_VERSION;
 #endif
@@ -1444,28 +1503,26 @@ main (argc, argv, envp)
 	  inhibit_warnings = 1;
 	else if (!strcmp (str, "W"))
 	  extra_warnings = 1;
-	else if (!strcmp (str, "Wimplicit"))
-	  warn_implicit = 1;
-	else if (!strcmp (str, "Wreturn-type"))
-	  warn_return_type = 1;
-	else if (!strcmp (str, "Wcomment"))
-	  ; /* cpp handles this one.  */
-	else if (!strcmp (str, "Wall"))
-	  {
-	    extra_warnings = 1;
-	    warn_implicit = 1;
-	    warn_return_type = 1;
-	  }
 	else if (!strcmp (str, "p"))
 	  profile_flag = 1;
+	else if (!strcmp (str, "gg"))
+	  write_symbols = GDB_DEBUG;
+#ifdef DBX_DEBUGGING_INFO
 	else if (!strcmp (str, "g"))
-	  write_symbols = 1;
+	  write_symbols = DBX_DEBUG;
 	else if (!strcmp (str, "G"))
-	  write_symbols = 2;
+	  write_symbols = DBX_DEBUG;
+#endif
+#ifdef SDB_DEBUGGING_INFO
+	else if (!strcmp (str, "g"))
+	  write_symbols = SDB_DEBUG;
+	else if (!strcmp (str, "G"))
+	  write_symbols = SDB_DEBUG;
+#endif
 	else if (!strcmp (str, "symout"))
 	  {
-	    if (write_symbols == 0)
-	      write_symbols = 1;
+	    if (write_symbols == NO_DEBUG)
+	      write_symbols = GDB_DEBUG;
 	    sym_file_name = argv[++i];
 	  }
 	else if (!strcmp (str, "o"))
@@ -1473,7 +1530,7 @@ main (argc, argv, envp)
 	    asm_file_name = argv[++i];
 	  }
 	else
-	  error ("Invalid switch, %s.", argv[i]);
+	  error ("Invalid option `%s'", argv[i]);
       }
     else
       filename = argv[i];
@@ -1489,6 +1546,9 @@ main (argc, argv, envp)
   OVERRIDE_OPTIONS;
 #endif
 
+  /* Now that register usage is specified, convert it to HARD_REG_SETs.  */
+  init_reg_sets_1 ();
+
   compile_file (filename);
 
 #ifndef USG
@@ -1509,6 +1569,8 @@ main (argc, argv, envp)
 
   if (errorcount)
     exit (FATAL_EXIT_CODE);
+  if (sorrycount)
+    exit (FATAL_EXIT_CODE);
   exit (SUCCESS_EXIT_CODE);
   return 34;
 }
@@ -1538,6 +1600,7 @@ set_target_switch (name)
 	  target_flags &= ~-target_switches[j].value;
 	else
 	  target_flags |= target_switches[j].value;
-	break;
+	return;
       }
+  error ("Invalid option `%s'", name);
 }
