@@ -67,6 +67,32 @@ void output_addressed_constants ();
 void output_constant ();
 void output_constructor ();
 
+static enum in_section {no_section, in_text, in_data} in_section = no_section;
+
+/* Tell assembler to switch to text section.  */
+
+void
+text_section ()
+{
+  if (in_section != in_text)
+    {
+      fprintf (asm_out_file, "%s\n", TEXT_SECTION_ASM_OP);
+      in_section = in_text;
+    }
+}
+
+/* Tell assembler to switch to data section.  */
+
+void
+data_section ()
+{
+  if (in_section != in_data)
+    {
+      fprintf (asm_out_file, "%s\n", DATA_SECTION_ASM_OP);
+      in_section = in_data;
+    }
+}
+
 /* Output a string of literal assembler code
    for an `asm' keyword used between functions.  */
 
@@ -104,14 +130,24 @@ assemble_function (decl)
 
   app_disable ();
 
-  /* Tell assembler to switch to text segment.  */
+  text_section ();
 
-  fprintf (asm_out_file, "%s\n", TEXT_SECTION_ASM_OP);
+#ifdef SDB_DEBUGGING_INFO
+  /* Make sure types are defined for debugger before fcn name is defined.  */
+  if (write_symbols == SDB_DEBUG)
+    sdbout_tags (gettags ());
+#endif
 
   /* Tell assembler to move to target machine's alignment for functions.  */
 
   ASM_OUTPUT_ALIGN (asm_out_file, floor_log2 (FUNCTION_BOUNDARY / BITS_PER_UNIT));
 
+#ifdef SDB_DEBUGGING_INFO
+  /* Output SDB definition of the function.  */
+  if (write_symbols == SDB_DEBUG)
+    sdbout_mark_begin_function ();
+#endif
+
   /* Make function name accessible from other files, if appropriate.  */
 
   if (TREE_PUBLIC (decl))
@@ -119,7 +155,7 @@ assemble_function (decl)
 
   /* Do any machine/system dependent processing of the function name */
 #ifdef ASM_DECLARE_FUNCTION_NAME
-  ASM_DECLARE_FUNCTION_NAME (asm_out_file, fnname);
+  ASM_DECLARE_FUNCTION_NAME (asm_out_file, fnname, current_function_decl);
 #else
   /* Standard thing is just output label for the function.  */
   ASM_OUTPUT_LABEL (asm_out_file, fnname);
@@ -155,7 +191,7 @@ make_function_rtl (decl)
 
    ASMSPEC is the user's specification of assembler symbol name to use.
    TOP_LEVEL is nonzero if this variable has file scope.
-   WRITE_SYMBOLS is 2 if writing dbx symbol output.
+   WRITE_SYMBOLS is DBX_DEBUG if writing dbx symbol output.
    The dbx data for a file-scope variable is written here.
    AT_END is nonzero if this is the special handling, at end of compilation,
    to define things that have had only tentative definitions.  */
@@ -165,7 +201,7 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
      tree decl;
      tree asmspec;
      int top_level;
-     int write_symbols;
+     enum debugger write_symbols;
      int at_end;
 {
   register char *name = IDENTIFIER_POINTER (DECL_NAME (decl));
@@ -175,7 +211,8 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
     {
       if (TREE_CODE (asmspec) != STRING_CST)
 	abort ();
-      name = (char *) oballoc (strlen (TREE_STRING_POINTER (asmspec)) + 2);
+      name = (char *) obstack_alloc (saveable_obstack,
+				     strlen (TREE_STRING_POINTER (asmspec)) + 2);
       name[0] = '*';
       strcpy (&name[1], TREE_STRING_POINTER (asmspec));
     }
@@ -198,7 +235,7 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
 	  char *label;
 
 	  ASM_FORMAT_PRIVATE_NAME (label, name, var_labelno);
-	  name = obstack_copy0 (current_obstack, label, strlen (label));
+	  name = obstack_copy0 (saveable_obstack, label, strlen (label));
 	  var_labelno++;
 	}
 
@@ -214,12 +251,6 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
 	   || TREE_CODE (TREE_TYPE (decl)) == UNION_TYPE);
     }
 
-  /* Output no assembler code for a function declaration.
-     Only definitions of functions output anything.  */
-
-  if (TREE_CODE (decl) == FUNCTION_DECL)
-    return;
-
   /* Normally no need to say anything for external references,
      since assembler considers all undefined symbols external.  */
 
@@ -231,6 +262,13 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
 #endif
       return;
     }
+
+  /* Output no assembler code for a function declaration.
+     Only definitions of functions output anything.  */
+
+  if (TREE_CODE (decl) == FUNCTION_DECL)
+    return;
+
   /* Don't output anything when a tentative file-scope definition is seen.
      But at end of compilation, do output code for them.  */
   if (! at_end && top_level
@@ -265,9 +303,16 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
 
   TREE_ASM_WRITTEN (decl) = 1;
 
-  if (write_symbols == 2)
+#ifdef DBX_DEBUGGING_INFO
+  /* File-scope global variables are output here.  */
+  if (write_symbols == DBX_DEBUG && top_level)
     dbxout_symbol (decl, 0);
-  else if (write_symbols != 0)
+#endif
+#ifdef SDB_DEBUGGING_INFO
+  if (write_symbols == SDB_DEBUG && top_level)
+    sdbout_symbol (decl, 0);
+#endif
+  if (write_symbols == GDB_DEBUG)
     /* Make sure the file is known to GDB even if it has no functions.  */
     set_current_gdbfile (DECL_SOURCE_FILE (decl));
 
@@ -288,13 +333,13 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
 		  / BITS_PER_UNIT);
       /* Round size up to multiple of BIGGEST_ALIGNMENT bits
 	 so that each uninitialized object starts on such a boundary.  */
-      size = ((size + (BIGGEST_ALIGNMENT / BITS_PER_UNIT) - 1)
-	      / (BIGGEST_ALIGNMENT / BITS_PER_UNIT)
-	      * (BIGGEST_ALIGNMENT / BITS_PER_UNIT));
+      int rounded = ((size + (BIGGEST_ALIGNMENT / BITS_PER_UNIT) - 1)
+		     / (BIGGEST_ALIGNMENT / BITS_PER_UNIT)
+		     * (BIGGEST_ALIGNMENT / BITS_PER_UNIT));
       if (TREE_PUBLIC (decl))
-	ASM_OUTPUT_COMMON (asm_out_file, name, size);
+	ASM_OUTPUT_COMMON (asm_out_file, name, size, rounded);
       else
-	ASM_OUTPUT_LOCAL (asm_out_file, name, size);
+	ASM_OUTPUT_LOCAL (asm_out_file, name, size, rounded);
       return;
     }
 
@@ -306,9 +351,9 @@ assemble_variable (decl, asmspec, top_level, write_symbols, at_end)
   output_addressed_constants (DECL_INITIAL (decl));
 
   if (TREE_READONLY (decl) && ! TREE_VOLATILE (decl))
-    fprintf (asm_out_file, "%s\n", TEXT_SECTION_ASM_OP);
+    text_section ();
   else
-    fprintf (asm_out_file, "%s\n", DATA_SECTION_ASM_OP);
+    data_section ();
 
   for (i = 0; DECL_ALIGN (decl) >= BITS_PER_UNIT << (i + 1); i++);
 
@@ -339,29 +384,20 @@ assemble_name (file, name)
    CONST_DOUBLE rtx's, and force those out to memory when necessary.  */
 
 /* Chain of all CONST_DOUBLE rtx's constructed for the current function.
-   They are chained through the third operand slot.  */
+   They are chained through the third operand slot, XEXP (r, 3).
+   A CONST_DOUBLE rtx has XEXP (r, 2) != cc0_rtx iff it is on this chain.
+   XEXP (r, 2) is either a MEM, or const0_rtx if no MEM has been made.  */
 
-extern rtx real_constant_chain;
+static rtx real_constant_chain;
 
-/* Return a CONST_DOUBLE for a specified `double' value
-   and machine mode.  */
+/* Return a CONST_DOUBLE for a value specified as a pair of ints.  */
 
 rtx
-immed_real_const_1 (d, mode)
-     double d;
+immed_double_const (i0, i1, mode)
+     int i0, i1;
      enum machine_mode mode;
-
 {
   register rtx r;
-  union {double d; int i[2];} u;
-  register int i0, i1;
-
-  /* Get the desired `double' value as two ints
-     since that is how they are stored in a CONST_DOUBLE.  */
-
-  u.d = d;
-  i0 = u.i[0];
-  i1 = u.i[1];
 
   /* Search the chain for an existing CONST_DOUBLE with the right value.
      If one is found, return it.  */
@@ -377,7 +413,7 @@ immed_real_const_1 (d, mode)
   XEXP (r, 3) = real_constant_chain;
   real_constant_chain = r;
 
-  /* Store const0_rtx in slot 2 just so most things won't barf.
+  /* Store const0_rtx in slot 2 since this CONST_DOUBLE is on the chain.
      Actual use of slot 2 is only through force_const_double_mem.  */
 
   XEXP (r, 2) = const0_rtx;
@@ -385,31 +421,50 @@ immed_real_const_1 (d, mode)
   return r;
 }
 
+/* Return a CONST_DOUBLE for a specified `double' value
+   and machine mode.  */
+
+rtx
+immed_real_const_1 (d, mode)
+     double d;
+     enum machine_mode mode;
+{
+  union {double d; int i[2];} u;
+
+  /* Get the desired `double' value as two ints
+     since that is how they are stored in a CONST_DOUBLE.  */
+
+  u.d = d;
+
+  return immed_double_const (u.i[0], u.i[1], mode);
+}
+
 /* Return a CONST_DOUBLE rtx for a value specified by EXP,
-   which must be a REAL_CST tree node.  Make only one CONST_DOUBLE
-   for each distinct value.  */
+   which must be a REAL_CST tree node.  */
 
 rtx
 immed_real_const (exp)
      tree exp;
 {
-  register rtx r;
-  r = immed_real_const_1 (TREE_REAL_CST (exp), TYPE_MODE (TREE_TYPE (exp)));
-
-  /* Associate exp and with this rtl value.  */
-  TREE_CST_RTL (exp) = r;
-
-  return r;
+  return immed_real_const_1 (TREE_REAL_CST (exp), TYPE_MODE (TREE_TYPE (exp)));
 }
 
 /* Given a CONST_DOUBLE, cause a constant in memory to be created
    (unless we already have one for the same value)
-   and return a MEM rtx to refer to it.  */
+   and return a MEM rtx to refer to it.
+   Put the CONST_DOUBLE on real_constant_chain if it isn't already there.  */
 
 rtx
 force_const_double_mem (r)
      rtx r;
 {
+  if (XEXP (r, 2) == cc0_rtx)
+    {
+      XEXP (r, 3) = real_constant_chain;
+      real_constant_chain = r;
+      XEXP (r, 2) = const0_rtx;
+    }
+
   if (XEXP (r, 2) == const0_rtx)
     {
       XEXP (r, 2) = force_const_mem (GET_MODE (r), r);
@@ -422,16 +477,22 @@ force_const_double_mem (r)
   return gen_rtx (MEM, GET_MODE (r), XEXP (XEXP (r, 2), 0));
 }
 
-/* At the start of a function, forget the memory-constants
-   previously made for CONST_DOUBLEs.  */
+/* At the end of a function, forget the memory-constants
+   previously made for CONST_DOUBLEs.  Mark them as not on real_constant_chain.
+   Also clear out real_constant_chain and clear out all the chain-pointers.  */
 
-static void
+void
 clear_const_double_mem ()
 {
-  register rtx r;
+  register rtx r, next;
 
-  for (r = real_constant_chain; r; r = XEXP (r, 3))
-    XEXP (r, 2) = 0;
+  for (r = real_constant_chain; r; r = next)
+    {
+      next = XEXP (r, 3);
+      XEXP (r, 3) = 0;
+      XEXP (r, 2) = cc0_rtx;
+    }
+  real_constant_chain = 0;
 }
 
 /* Given an expression EXP with a constant value,
@@ -703,7 +764,7 @@ record_constant_1 (exp)
   register int len;
   register enum tree_code code = TREE_CODE (exp);
 
-  obstack_1grow (&permanent_obstack, (unsigned char) code);
+  obstack_1grow (&permanent_obstack, (unsigned int) code);
 
   if (code == INTEGER_CST)
     {
@@ -750,14 +811,14 @@ record_constant_1 (exp)
     }
   else if (code == PLUS_EXPR || code == MINUS_EXPR)
     {
-      obstack_1grow (&permanent_obstack, (char) code);
+      obstack_1grow (&permanent_obstack, (int) code);
       record_constant_1 (TREE_OPERAND (exp, 0));
       record_constant_1 (TREE_OPERAND (exp, 1));
       return;
     }
   else if (code == NOP_EXPR || code == CONVERT_EXPR)
     {
-      obstack_1grow (&permanent_obstack, (char) code);
+      obstack_1grow (&permanent_obstack, (int) code);
       record_constant_1 (TREE_OPERAND (exp, 0));
       return;
     }
@@ -806,11 +867,11 @@ get_or_assign_label (exp)
   /* Now output assembler code to define that label
      and follow it with the data of EXP.  */
 
-  /* First switch to text segment, except for writable strings.  */
+  /* First switch to text section, except for writable strings.  */
   if ((TREE_CODE (exp) == STRING_CST) && flag_writable_strings)
-    fprintf (asm_out_file, "%s\n", DATA_SECTION_ASM_OP);
+    data_section ();
   else
-    fprintf (asm_out_file, "%s\n", TEXT_SECTION_ASM_OP);
+    text_section ();
 
   /* Align the location counter as required by EXP's data type.  */
   for (i = 0; TYPE_ALIGN (TREE_TYPE (exp)) >= BITS_PER_UNIT << (i + 1); i++);
@@ -850,15 +911,24 @@ output_constant_def (exp)
   register rtx def;
   char label[10];
 
+  if (TREE_CODE (exp) == INTEGER_CST)
+    abort ();			/* No TREE_CST_RTL slot in these.  */
+
   if (TREE_CST_RTL (exp))
     return TREE_CST_RTL (exp);
 
+  if (TREE_PERMANENT (exp))
+    end_temporary_allocation ();
+
   def = gen_rtx (SYMBOL_REF, Pmode, get_or_assign_label (exp));
 
   TREE_CST_RTL (exp)
     = gen_rtx (MEM, TYPE_MODE (TREE_TYPE (exp)), def);
   TREE_CST_RTL (exp)->unchanging = 1;
 
+  if (TREE_PERMANENT (exp))
+    resume_temporary_allocation ();
+
   return TREE_CST_RTL (exp);
 }
 
@@ -877,7 +947,6 @@ void
 init_const_rtx_hash_table ()
 {
   bzero (const_rtx_hash_table, sizeof const_rtx_hash_table);
-  clear_const_double_mem ();
 }
 
 struct rtx_const
@@ -895,12 +964,13 @@ struct rtx_const
    They are stored into VALUE.  */
 
 static void
-decode_rtx_const (x, value)
+decode_rtx_const (mode, x, value)
+     enum machine_mode mode;
      rtx x;
      struct rtx_const *value;
 {
   value->kind = RTX_INT;	/* Most usual kind. */
-  value->mode = GET_MODE (x);
+  value->mode = mode;
   value->un.addr.base = 0;
   value->un.addr.offset = 0;
 
@@ -956,14 +1026,15 @@ decode_rtx_const (x, value)
 /* Compute a hash code for a constant RTL expression.  */
 
 int
-const_hash_rtx (x)
+const_hash_rtx (mode, x)
+     enum machine_mode mode;
      rtx x;
 {
   register int hi, i, len;
   register char *p;
 
   struct rtx_const value;
-  decode_rtx_const (x, &value);
+  decode_rtx_const (mode, x, &value);
 
   /* Compute hashing function */
   hi = 0;
@@ -979,7 +1050,8 @@ const_hash_rtx (x)
    Return 1 if DESC describes a constant with the same value as X.  */
 
 static int
-compare_constant_rtx (x, desc)
+compare_constant_rtx (mode, x, desc)
+     enum machine_mode mode;
      rtx x;
      struct constant_descriptor *desc;
 {
@@ -988,7 +1060,7 @@ compare_constant_rtx (x, desc)
   register int len;
   struct rtx_const value;
 
-  decode_rtx_const (x, &value);
+  decode_rtx_const (mode, x, &value);
   strp = (int *) &value;
   len = sizeof value / sizeof (int);
 
@@ -1004,14 +1076,15 @@ compare_constant_rtx (x, desc)
    It is up to the caller to enter the descriptor in the hash table.  */
 
 static struct constant_descriptor *
-record_constant_rtx (x)
+record_constant_rtx (mode, x)
+     enum machine_mode mode;
      rtx x;
 {
   struct constant_descriptor *ptr = 0;
   int buf;
   struct rtx_const value;
 
-  decode_rtx_const (x, &value);
+  decode_rtx_const (mode, x, &value);
 
   obstack_grow (saveable_obstack, &ptr, sizeof ptr);
   obstack_grow (saveable_obstack, &buf, sizeof buf);
@@ -1040,10 +1113,10 @@ force_const_mem (mode, x)
      to see if any of them describes X.  If yes, the descriptor records
      the label number already assigned.  */
 
-  hash = const_hash_rtx (x);
+  hash = const_hash_rtx (mode, x);
 
   for (desc = const_rtx_hash_table[hash]; desc; desc = desc->next)
-    if (compare_constant_rtx (x, desc))
+    if (compare_constant_rtx (mode, x, desc))
       {
 	found = desc->label;
 	break;
@@ -1051,23 +1124,29 @@ force_const_mem (mode, x)
 
   if (found == 0)
     {
+      int align;
+
       /* No constant equal to X is known to have been output.
 	 Make a constant descriptor to enter X in the hash table.
 	 Assign the label number and record it in the descriptor for
 	 future calls to this function to find.  */
 
-      desc = record_constant_rtx (x);
+      desc = record_constant_rtx (mode, x);
       desc->next = const_rtx_hash_table[hash];
       const_rtx_hash_table[hash] = desc;
 
       /* Now output assembler code to define that label
 	 and follow it with the data of EXP.  */
 
-      /* First switch to text segment.  */
-      fprintf (asm_out_file, "%s\n", TEXT_SECTION_ASM_OP);
+      /* First switch to text section.  */
+      text_section ();
 
       /* Align the location counter as required by EXP's data type.  */
-      ASM_OUTPUT_ALIGN (asm_out_file, exact_log2 (UNITS_PER_WORD));
+      align = (mode == VOIDmode) ? UNITS_PER_WORD : GET_MODE_SIZE (mode);
+      if (align > BIGGEST_ALIGNMENT / BITS_PER_UNIT)
+	align = BIGGEST_ALIGNMENT / BITS_PER_UNIT;
+
+      ASM_OUTPUT_ALIGN (asm_out_file, exact_log2 (align));
 
       /* Output the label itself.  */
       ASM_OUTPUT_INTERNAL_LABEL (asm_out_file, "LC", const_labelno);
@@ -1085,7 +1164,20 @@ force_const_mem (mode, x)
 	    ASM_OUTPUT_FLOAT (asm_out_file, u.d);
 	}
       else
-	ASM_OUTPUT_INT (asm_out_file, x);
+	switch (mode)
+	  {
+	  case SImode:
+	    ASM_OUTPUT_INT (asm_out_file, x);
+	    break;
+
+	  case HImode:
+	    ASM_OUTPUT_SHORT (asm_out_file, x);
+	    break;
+
+	  case QImode:
+	    ASM_OUTPUT_CHAR (asm_out_file, x);
+	    break;
+	  }
 
       /* Create a string containing the label name, in LABEL.  */
       ASM_GENERATE_INTERNAL_LABEL (label, "LC", const_labelno);
@@ -1194,6 +1286,37 @@ output_constant (exp, size)
     case INTEGER_TYPE:
     case ENUMERAL_TYPE:
     case POINTER_TYPE:
+      while (TREE_CODE (exp) == NOP_EXPR || TREE_CODE (exp) == CONVERT_EXPR)
+	exp = TREE_OPERAND (exp, 0);
+
+      if (TYPE_MODE (TREE_TYPE (exp)) == DImode)
+	{
+	  if (TREE_CODE (exp) == INTEGER_CST)
+	    {
+#ifdef WORDS_BIG_ENDIAN
+	      ASM_OUTPUT_INT (asm_out_file,
+			      gen_rtx (CONST_INT, VOIDmode,
+				       TREE_INT_CST_LOW (exp)));
+	      ASM_OUTPUT_INT (asm_out_file,
+			      gen_rtx (CONST_INT, VOIDmode,
+				       TREE_INT_CST_HIGH (exp)));
+#else
+	      ASM_OUTPUT_INT (asm_out_file,
+			      gen_rtx (CONST_INT, VOIDmode,
+				       TREE_INT_CST_HIGH (exp)));
+	      ASM_OUTPUT_INT (asm_out_file,
+			      gen_rtx (CONST_INT, VOIDmode,
+				       TREE_INT_CST_LOW (exp)));
+#endif
+	      size -= 8;
+	      break;
+	    }
+	  else
+	    error ("8-byte integer constant expression too complicated");
+
+	  break;
+	}
+
       x = expand_expr (exp, 0, VOIDmode, EXPAND_SUM);
 
       if (size == 1)
